import pytest

import salt.modules.cmdmod
import salt.modules.pkg_resource
import salt.modules.yumpkg
import salt.utils.pkg.rpm


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        salt.modules.yumpkg: {
            "__salt__": {
                "cmd.run": salt.modules.cmdmod.run,
                "pkg_resource.add_pkg": salt.modules.pkg_resource.add_pkg,
                "pkg_resource.format_pkg_list": salt.modules.pkg_resource.format_pkg_list,
            },
            "__grains__": {"osarch": salt.utils.pkg.rpm.get_osarch()},
        },
    }


@pytest.mark.slow_test
def test_yum_list_pkgs(grains):
    """
    compare the output of rpm -qa vs the return of yumpkg.list_pkgs,
    make sure that any changes to ympkg.list_pkgs still returns.
    """

    if grains["os_family"] != "RedHat":
        pytest.skip("Skip if not RedHat")
    cmd = [
        "rpm",
        "-qa",
        "--queryformat",
        "%{NAME}\n",
    ]
    known_pkgs = salt.modules.cmdmod.run(cmd, python_shell=False)
    listed_pkgs = salt.modules.yumpkg.list_pkgs()
    for line in known_pkgs.splitlines():
        assert any(line in d for d in listed_pkgs)



import os

import pytest

import salt.crypt
from tests.support.mock import MagicMock, MockCall, mock_open, patch


@pytest.mark.slow_test
def test_gen_keys():
    open_priv_wb = MockCall(f"/keydir{os.sep}keyname.pem", "wb+")
    open_pub_wb = MockCall(f"/keydir{os.sep}keyname.pub", "wb+")

    with patch.multiple(
        os,
        umask=MagicMock(),
        chmod=MagicMock(),
        access=MagicMock(return_value=True),
    ):
        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
            "os.path.isfile", return_value=True
        ):
            result = salt.crypt.gen_keys("/keydir", "keyname", 2048)
            assert result == f"/keydir{os.sep}keyname.pem", result
            assert open_priv_wb not in m_open.calls
            assert open_pub_wb not in m_open.calls

        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
            "os.path.isfile", return_value=False
        ):
            salt.crypt.gen_keys("/keydir", "keyname", 2048)
            assert open_priv_wb in m_open.calls
            assert open_pub_wb in m_open.calls


@patch("os.umask", MagicMock())
@patch("os.chmod", MagicMock())
@patch("os.chown", MagicMock(), create=True)
@patch("os.access", MagicMock(return_value=True))
@pytest.mark.slow_test
def test_gen_keys_with_passphrase():
    key_path = os.path.join(os.sep, "keydir")
    open_priv_wb = MockCall(os.path.join(key_path, "keyname.pem"), "wb+")
    open_pub_wb = MockCall(os.path.join(key_path, "keyname.pub"), "wb+")

    with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
        "os.path.isfile", return_value=True
    ):
        assert salt.crypt.gen_keys(
            key_path, "keyname", 2048, passphrase="password"
        ) == os.path.join(key_path, "keyname.pem")
        result = salt.crypt.gen_keys(key_path, "keyname", 2048, passphrase="password")
        assert result == os.path.join(key_path, "keyname.pem"), result
        assert open_priv_wb not in m_open.calls
        assert open_pub_wb not in m_open.calls

    with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
        "os.path.isfile", return_value=False
    ):
        salt.crypt.gen_keys(key_path, "keyname", 2048)
        assert open_priv_wb in m_open.calls
        assert open_pub_wb in m_open.calls



"""
:codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.sysmod as sysmod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class MockDocstringable:
    def __init__(self, docstr):
        self.__doc__ = docstr

    def set_module_docstring(self, docstr):
        self.__globals__ = {"__doc__": docstr}


class Mockstate:
    """
    Mock of State
    """

    class State:
        """
        Mock state functions
        """

        states = {}

        def __init__(self, opts):
            pass


class Mockrunner:
    """
    Mock of runner
    """

    class Runner:
        """
        Mock runner functions
        """

        def __init__(self, opts):
            pass

        @property
        def functions(self):
            return sysmod.__salt__


class Mockloader:
    """
    Mock of loader
    """

    functions = []  # ? does not have any effect on existing tests

    def __init__(self):
        pass

    def returners(self, opts, lst):
        """
        Mock returner functions
        """
        return sysmod.__salt__

    def render(self, opts, lst):
        """
        Mock renderers
        """
        return (
            sysmod.__salt__
        )  # renderers do not have '.'s; but whatever. This is for convenience


class SysmodTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.sysmod
    """

    def setup_loader_modules(self):
        return {sysmod: {"__salt__": self.salt_dunder}}

    @classmethod
    def setUpClass(cls):
        cls._modules = set()
        cls._functions = [
            "exist.exist",
            "sys.doc",
            "sys.list_functions",
            "sys.list_modules",
            "sysctl.get",
            "sysctl.show",
            "system.halt",
            "system.reboot",
            "udev.name",
            "udev.path",
            "user.add",
            "user.info",
            "user.rename",
        ]
        cls._docstrings = {}
        cls._statedocstrings = {}

        cls.salt_dunder = {}

        for func in cls._functions:
            docstring = f"docstring for {func}"

            cls.salt_dunder[func] = MockDocstringable(docstring)
            cls._docstrings[func] = docstring

            module = func.split(".", maxsplit=1)[0]
            cls._statedocstrings[func] = docstring
            cls._statedocstrings[module] = f"docstring for {module}"

            cls._modules.add(func.split(".", maxsplit=1)[0])

            docstring = f"docstring for {func}"
            mock = MockDocstringable(docstring)
            mock.set_module_docstring(
                "docstring for {}".format(func.split(".", maxsplit=1)[0])
            )
            Mockstate.State.states[func] = mock

        cls._modules = sorted(list(cls._modules))

        cls.state_patcher = patch("salt.state", Mockstate())
        cls.state_patcher.start()
        cls.runner_patcher = patch("salt.runner", Mockrunner())
        cls.runner_patcher.start()
        cls.loader_patcher = patch("salt.loader", Mockloader())
        cls.loader_patcher.start()

    # 'doc' function tests: 2

    @classmethod
    def tearDownClass(cls):
        cls.runner_patcher.stop()
        cls.state_patcher.stop()
        cls.loader_patcher.stop()
        for attrname in (
            "_modules",
            "_functions",
            "_docstrings",
            "_statedocstrings",
            "salt_dunder",
            "runner_patcher",
            "state_patcher",
            "loader_patcher",
        ):
            try:
                delattr(cls, attrname)
            except AttributeError:
                continue

    def test_doc(self):
        """
        Test if it returns the docstrings for all modules.
        """
        self.assertDictEqual(sysmod.doc(), self._docstrings)

        self.assertDictEqual(
            sysmod.doc("sys.doc"), {"sys.doc": "docstring for sys.doc"}
        )

    # 'state_doc' function tests: 2

    def test_state_doc(self):
        """
        Test if it returns the docstrings for all states.
        """
        self.assertDictEqual(sysmod.state_doc(), self._statedocstrings)

        self.assertDictEqual(
            sysmod.state_doc("sys.doc"),
            {"sys": "docstring for sys", "sys.doc": "docstring for sys.doc"},
        )

    # 'runner_doc' function tests: 2

    def test_runner_doc(self):
        """
        Test if it returns the docstrings for all runners.
        """
        self.assertDictEqual(sysmod.runner_doc(), self._docstrings)

        self.assertDictEqual(
            sysmod.runner_doc("sys.doc"), {"sys.doc": "docstring for sys.doc"}
        )

    # 'returner_doc' function tests: 2

    def test_returner_doc(self):
        """
        Test if it returns the docstrings for all returners.
        """
        self.assertDictEqual(sysmod.returner_doc(), self._docstrings)

        self.assertDictEqual(
            sysmod.returner_doc("sys.doc"), {"sys.doc": "docstring for sys.doc"}
        )

    # 'renderer_doc' function tests: 2

    def test_renderer_doc(self):
        """
        Test if it returns the docstrings for all renderers.
        """
        self.assertDictEqual(sysmod.renderer_doc(), self._docstrings)

        self.assertDictEqual(
            sysmod.renderer_doc("sys.doc"), {"sys.doc": "docstring for sys.doc"}
        )

    # 'list_functions' function tests: 7

    def test_list_functions(self):
        """
        Test if it lists the functions for all modules.
        """
        self.assertListEqual(sysmod.list_functions(), self._functions)

        self.assertListEqual(sysmod.list_functions("nonexist"), [])

        # list all functions in/given a specific module
        self.assertListEqual(
            sysmod.list_functions("sys"),
            ["sys.doc", "sys.list_functions", "sys.list_modules"],
        )

        # globs can be used for both module names and function names:
        self.assertListEqual(
            sysmod.list_functions("sys*"),
            [
                "sys.doc",
                "sys.list_functions",
                "sys.list_modules",
                "sysctl.get",
                "sysctl.show",
                "system.halt",
                "system.reboot",
            ],
        )
        self.assertListEqual(
            sysmod.list_functions("sys.list*"),
            ["sys.list_functions", "sys.list_modules"],
        )

        # "list", or check for a specific function:
        self.assertListEqual(sysmod.list_functions("sys.list"), [])
        self.assertListEqual(sysmod.list_functions("exist.exist"), ["exist.exist"])

    # 'list_modules' function tests: 4

    def test_list_modules(self):
        """
        Test if it lists the modules loaded on the minion
        """
        self.assertListEqual(sysmod.list_modules(), self._modules)

        self.assertListEqual(sysmod.list_modules("nonexist"), [])

        self.assertListEqual(sysmod.list_modules("user"), ["user"])

        self.assertListEqual(sysmod.list_modules("s*"), ["sys", "sysctl", "system"])

    # 'reload_modules' function tests: 1

    def test_reload_modules(self):
        """
        Test if it tell the minion to reload the execution modules
        """
        self.assertTrue(sysmod.reload_modules())

    # 'argspec' function tests: 1

    def test_argspec(self):
        """
        Test if it return the argument specification
        of functions in Salt execution modules.
        """
        self.assertDictEqual(sysmod.argspec(), {})

    # 'state_argspec' function tests: 1

    def test_state_argspec(self):
        """
        Test if it return the argument specification
        of functions in Salt state modules.
        """
        self.assertDictEqual(sysmod.state_argspec(), {})

    # 'returner_argspec' function tests: 1

    def test_returner_argspec(self):
        """
        Test if it return the argument specification
        of functions in Salt returner modules.
        """
        self.assertDictEqual(sysmod.returner_argspec(), {})

    # 'runner_argspec' function tests: 1

    def test_runner_argspec(self):
        """
        Test if it return the argument specification of functions in Salt runner
        modules.
        """
        self.assertDictEqual(sysmod.runner_argspec(), {})

    # 'list_state_functions' function tests: 7

    def test_list_state_functions(self):
        """
        Test if it lists the functions for all state modules.
        """
        self.assertListEqual(sysmod.list_state_functions(), self._functions)

        self.assertListEqual(sysmod.list_state_functions("nonexist"), [])

        # list all functions in/given a specific module
        self.assertListEqual(
            sysmod.list_state_functions("sys"),
            ["sys.doc", "sys.list_functions", "sys.list_modules"],
        )

        # globs can be used for both module names and function names:
        self.assertListEqual(
            sysmod.list_state_functions("sys*"),
            [
                "sys.doc",
                "sys.list_functions",
                "sys.list_modules",
                "sysctl.get",
                "sysctl.show",
                "system.halt",
                "system.reboot",
            ],
        )
        self.assertListEqual(
            sysmod.list_state_functions("sys.list*"),
            ["sys.list_functions", "sys.list_modules"],
        )

        # "list", or check for a specific function:
        self.assertListEqual(sysmod.list_state_functions("sys.list"), [])
        self.assertListEqual(
            sysmod.list_state_functions("exist.exist"), ["exist.exist"]
        )

    # 'list_state_modules' function tests: 4

    def test_list_state_modules(self):
        """
        Test if it lists the modules loaded on the minion.
        """
        self.assertListEqual(sysmod.list_state_modules(), self._modules)

        self.assertListEqual(sysmod.list_state_modules("nonexist"), [])

        self.assertListEqual(sysmod.list_state_modules("user"), ["user"])

        self.assertListEqual(
            sysmod.list_state_modules("s*"), ["sys", "sysctl", "system"]
        )

    # 'list_runners' function tests: 4

    def test_list_runners(self):
        """
        Test if it list the runners loaded on the minion.
        """
        self.assertListEqual(sysmod.list_runners(), self._modules)

        self.assertListEqual(sysmod.list_runners("nonexist"), [])

        self.assertListEqual(sysmod.list_runners("user"), ["user"])

        self.assertListEqual(sysmod.list_runners("s*"), ["sys", "sysctl", "system"])

    # 'list_runner_functions' function tests: 7

    def test_list_runner_functions(self):
        """
        Test if it lists the functions for all runner modules.
        """
        self.assertListEqual(sysmod.list_runner_functions(), self._functions)

        self.assertListEqual(sysmod.list_runner_functions("nonexist"), [])

        # list all functions in/given a specific module
        self.assertListEqual(
            sysmod.list_runner_functions("sys"),
            ["sys.doc", "sys.list_functions", "sys.list_modules"],
        )

        # globs can be used for both module names and function names:
        self.assertListEqual(
            sysmod.list_runner_functions("sys*"),
            [
                "sys.doc",
                "sys.list_functions",
                "sys.list_modules",
                "sysctl.get",
                "sysctl.show",
                "system.halt",
                "system.reboot",
            ],
        )
        self.assertListEqual(
            sysmod.list_runner_functions("sys.list*"),
            ["sys.list_functions", "sys.list_modules"],
        )

        # "list", or check for a specific function:
        self.assertListEqual(sysmod.list_runner_functions("sys.list"), [])
        self.assertListEqual(
            sysmod.list_runner_functions("exist.exist"), ["exist.exist"]
        )

    # 'list_returners' function tests: 4

    def test_list_returners(self):
        """
        Test if it lists the returners loaded on the minion
        """
        self.assertListEqual(sysmod.list_returners(), self._modules)

        self.assertListEqual(sysmod.list_returners("nonexist"), [])

        self.assertListEqual(sysmod.list_returners("user"), ["user"])

        self.assertListEqual(sysmod.list_returners("s*"), ["sys", "sysctl", "system"])

    # 'list_returner_functions' function tests: 7

    def test_list_returner_functions(self):
        """
        Test if it lists the functions for all returner modules.
        """
        self.assertListEqual(sysmod.list_returner_functions(), self._functions)

        self.assertListEqual(sysmod.list_returner_functions("nonexist"), [])

        # list all functions in/given a specific module
        self.assertListEqual(
            sysmod.list_returner_functions("sys"),
            ["sys.doc", "sys.list_functions", "sys.list_modules"],
        )

        # globs can be used for both module names and function names:
        self.assertListEqual(
            sysmod.list_returner_functions("sys*"),
            [
                "sys.doc",
                "sys.list_functions",
                "sys.list_modules",
                "sysctl.get",
                "sysctl.show",
                "system.halt",
                "system.reboot",
            ],
        )
        self.assertListEqual(
            sysmod.list_returner_functions("sys.list*"),
            ["sys.list_functions", "sys.list_modules"],
        )

        # "list", or check for a specific function:
        self.assertListEqual(sysmod.list_returner_functions("sys.list"), [])
        self.assertListEqual(
            sysmod.list_returner_functions("exist.exist"), ["exist.exist"]
        )

    # 'list_renderers' function tests: 4

    def test_list_renderers(self):
        """
        Test if it list the renderers loaded on the minion.
        """
        self.assertListEqual(sysmod.list_renderers(), self._functions)

        self.assertListEqual(sysmod.list_renderers("nonexist"), [])

        self.assertListEqual(sysmod.list_renderers("user.info"), ["user.info"])

        self.assertListEqual(
            sysmod.list_renderers("syst*"), ["system.halt", "system.reboot"]
        )



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A simple template system that compiles templates to Python code.

Basic usage looks like::

    t = template.Template("<html>{{ myvalue }}</html>")
    print(t.generate(myvalue="XXX"))

`Loader` is a class that loads templates from a root directory and caches
the compiled templates::

    loader = template.Loader("/home/btaylor")
    print(loader.load("test.html").generate(myvalue="XXX"))

We compile all templates to raw Python. Error-reporting is currently... uh,
interesting. Syntax for the templates::

    ### base.html
    <html>
      <head>
        <title>{% block title %}Default title{% end %}</title>
      </head>
      <body>
        <ul>
          {% for student in students %}
            {% block student %}
              <li>{{ escape(student.name) }}</li>
            {% end %}
          {% end %}
        </ul>
      </body>
    </html>

    ### bold.html
    {% extends "base.html" %}

    {% block title %}A bolder title{% end %}

    {% block student %}
      <li><span style="bold">{{ escape(student.name) }}</span></li>
    {% end %}

Unlike most other template systems, we do not put any restrictions on the
expressions you can include in your statements. ``if`` and ``for`` blocks get
translated exactly into Python, so you can do complex expressions like::

   {% for student in [p for p in people if p.student and p.age > 23] %}
     <li>{{ escape(student.name) }}</li>
   {% end %}

Translating directly to Python means you can apply functions to expressions
easily, like the ``escape()`` function in the examples above. You can pass
functions in to your template just like any other variable
(In a `.RequestHandler`, override `.RequestHandler.get_template_namespace`)::

   ### Python code
   def add(x, y):
      return x + y
   template.execute(add=add)

   ### The template
   {{ add(1, 2) }}

We provide the functions `escape() <.xhtml_escape>`, `.url_escape()`,
`.json_encode()`, and `.squeeze()` to all templates by default.

Typical applications do not create `Template` or `Loader` instances by
hand, but instead use the `~.RequestHandler.render` and
`~.RequestHandler.render_string` methods of
`tornado.web.RequestHandler`, which load templates automatically based
on the ``template_path`` `.Application` setting.

Variable names beginning with ``_tt_`` are reserved by the template
system and should not be used by application code.

Syntax Reference
----------------

Template expressions are surrounded by double curly braces: ``{{ ... }}``.
The contents may be any python expression, which will be escaped according
to the current autoescape setting and inserted into the output.  Other
template directives use ``{% %}``.

To comment out a section so that it is omitted from the output, surround it
with ``{# ... #}``.

These tags may be escaped as ``{{!``, ``{%!``, and ``{#!``
if you need to include a literal ``{{``, ``{%``, or ``{#`` in the output.


``{% apply *function* %}...{% end %}``
    Applies a function to the output of all template code between ``apply``
    and ``end``::

        {% apply linkify %}{{name}} said: {{message}}{% end %}

    Note that as an implementation detail apply blocks are implemented
    as nested functions and thus may interact strangely with variables
    set via ``{% set %}``, or the use of ``{% break %}`` or ``{% continue %}``
    within loops.

``{% autoescape *function* %}``
    Sets the autoescape mode for the current file.  This does not affect
    other files, even those referenced by ``{% include %}``.  Note that
    autoescaping can also be configured globally, at the `.Application`
    or `Loader`.::

        {% autoescape xhtml_escape %}
        {% autoescape None %}

``{% block *name* %}...{% end %}``
    Indicates a named, replaceable block for use with ``{% extends %}``.
    Blocks in the parent template will be replaced with the contents of
    the same-named block in a child template.::

        <!-- base.html -->
        <title>{% block title %}Default title{% end %}</title>

        <!-- mypage.html -->
        {% extends "base.html" %}
        {% block title %}My page title{% end %}

``{% comment ... %}``
    A comment which will be removed from the template output.  Note that
    there is no ``{% end %}`` tag; the comment goes from the word ``comment``
    to the closing ``%}`` tag.

``{% extends *filename* %}``
    Inherit from another template.  Templates that use ``extends`` should
    contain one or more ``block`` tags to replace content from the parent
    template.  Anything in the child template not contained in a ``block``
    tag will be ignored.  For an example, see the ``{% block %}`` tag.

``{% for *var* in *expr* %}...{% end %}``
    Same as the python ``for`` statement.  ``{% break %}`` and
    ``{% continue %}`` may be used inside the loop.

``{% from *x* import *y* %}``
    Same as the python ``import`` statement.

``{% if *condition* %}...{% elif *condition* %}...{% else %}...{% end %}``
    Conditional statement - outputs the first section whose condition is
    true.  (The ``elif`` and ``else`` sections are optional)

``{% import *module* %}``
    Same as the python ``import`` statement.

``{% include *filename* %}``
    Includes another template file.  The included file can see all the local
    variables as if it were copied directly to the point of the ``include``
    directive (the ``{% autoescape %}`` directive is an exception).
    Alternately, ``{% module Template(filename, **kwargs) %}`` may be used
    to include another template with an isolated namespace.

``{% module *expr* %}``
    Renders a `~tornado.web.UIModule`.  The output of the ``UIModule`` is
    not escaped::

        {% module Template("foo.html", arg=42) %}

    ``UIModules`` are a feature of the `tornado.web.RequestHandler`
    class (and specifically its ``render`` method) and will not work
    when the template system is used on its own in other contexts.

``{% raw *expr* %}``
    Outputs the result of the given expression without autoescaping.

``{% set *x* = *y* %}``
    Sets a local variable.

``{% try %}...{% except %}...{% else %}...{% finally %}...{% end %}``
    Same as the python ``try`` statement.

``{% while *condition* %}... {% end %}``
    Same as the python ``while`` statement.  ``{% break %}`` and
    ``{% continue %}`` may be used inside the loop.

``{% whitespace *mode* %}``
    Sets the whitespace mode for the remainder of the current file
    (or until the next ``{% whitespace %}`` directive). See
    `filter_whitespace` for available options. New in Tornado 4.3.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import datetime
import linecache
import os.path
import posixpath
import re
import threading

from salt.ext.tornado import escape
from salt.ext.tornado.log import app_log
from salt.ext.tornado.util import ObjectDict, exec_in, unicode_type, PY3

if PY3:
    from io import StringIO
else:
    from cStringIO import StringIO

_DEFAULT_AUTOESCAPE = "xhtml_escape"
_UNSET = object()


def filter_whitespace(mode, text):
    """Transform whitespace in ``text`` according to ``mode``.

    Available modes are:

    * ``all``: Return all whitespace unmodified.
    * ``single``: Collapse consecutive whitespace with a single whitespace
      character, preserving newlines.
    * ``oneline``: Collapse all runs of whitespace into a single space
      character, removing all newlines in the process.

    .. versionadded:: 4.3
    """
    if mode == 'all':
        return text
    elif mode == 'single':
        text = re.sub(r"([\t ]+)", " ", text)
        text = re.sub(r"(\s*\n\s*)", "\n", text)
        return text
    elif mode == 'oneline':
        return re.sub(r"(\s+)", " ", text)
    else:
        raise Exception("invalid whitespace mode %s" % mode)


class Template(object):
    """A compiled template.

    We compile into Python from the given template_string. You can generate
    the template from variables with generate().
    """
    # note that the constructor's signature is not extracted with
    # autodoc because _UNSET looks like garbage.  When changing
    # this signature update website/sphinx/template.rst too.
    def __init__(self, template_string, name="<string>", loader=None,
                 compress_whitespace=_UNSET, autoescape=_UNSET,
                 whitespace=None):
        """Construct a Template.

        :arg str template_string: the contents of the template file.
        :arg str name: the filename from which the template was loaded
            (used for error message).
        :arg tornado.template.BaseLoader loader: the `~tornado.template.BaseLoader` responsible for this template,
            used to resolve ``{% include %}`` and ``{% extend %}``
            directives.
        :arg bool compress_whitespace: Deprecated since Tornado 4.3.
            Equivalent to ``whitespace="single"`` if true and
            ``whitespace="all"`` if false.
        :arg str autoescape: The name of a function in the template
            namespace, or ``None`` to disable escaping by default.
        :arg str whitespace: A string specifying treatment of whitespace;
            see `filter_whitespace` for options.

        .. versionchanged:: 4.3
           Added ``whitespace`` parameter; deprecated ``compress_whitespace``.
        """
        self.name = escape.native_str(name)

        if compress_whitespace is not _UNSET:
            # Convert deprecated compress_whitespace (bool) to whitespace (str).
            if whitespace is not None:
                raise Exception("cannot set both whitespace and compress_whitespace")
            whitespace = "single" if compress_whitespace else "all"
        if whitespace is None:
            if loader and loader.whitespace:
                whitespace = loader.whitespace
            else:
                # Whitespace defaults by filename.
                if name.endswith(".html") or name.endswith(".js"):
                    whitespace = "single"
                else:
                    whitespace = "all"
        # Validate the whitespace setting.
        filter_whitespace(whitespace, '')

        if autoescape is not _UNSET:
            self.autoescape = autoescape
        elif loader:
            self.autoescape = loader.autoescape
        else:
            self.autoescape = _DEFAULT_AUTOESCAPE

        self.namespace = loader.namespace if loader else {}
        reader = _TemplateReader(name, escape.native_str(template_string),
                                 whitespace)
        self.file = _File(self, _parse(reader, self))
        self.code = self._generate_python(loader)
        self.loader = loader
        try:
            # Under python2.5, the fake filename used here must match
            # the module name used in __name__ below.
            # The dont_inherit flag prevents template.py's future imports
            # from being applied to the generated code.
            self.compiled = compile(
                escape.to_unicode(self.code),
                "%s.generated.py" % self.name.replace('.', '_'),
                "exec", dont_inherit=True)
        except Exception:
            formatted_code = _format_code(self.code).rstrip()
            app_log.error("%s code:\n%s", self.name, formatted_code)
            raise

    def generate(self, **kwargs):
        """Generate this template with the given arguments."""
        namespace = {
            "escape": escape.xhtml_escape,
            "xhtml_escape": escape.xhtml_escape,
            "url_escape": escape.url_escape,
            "json_encode": escape.json_encode,
            "squeeze": escape.squeeze,
            "linkify": escape.linkify,
            "datetime": datetime,
            "_tt_utf8": escape.utf8,  # for internal use
            "_tt_string_types": (unicode_type, bytes),
            # __name__ and __loader__ allow the traceback mechanism to find
            # the generated source code.
            "__name__": self.name.replace('.', '_'),
            "__loader__": ObjectDict(get_source=lambda name: self.code),
        }
        namespace.update(self.namespace)
        namespace.update(kwargs)
        exec_in(self.compiled, namespace)
        execute = namespace["_tt_execute"]
        # Clear the traceback module's cache of source data now that
        # we've generated a new template (mainly for this module's
        # unittests, where different tests reuse the same name).
        linecache.clearcache()
        return execute()

    def _generate_python(self, loader):
        buffer = StringIO()
        try:
            # named_blocks maps from names to _NamedBlock objects
            named_blocks = {}
            ancestors = self._get_ancestors(loader)
            ancestors.reverse()
            for ancestor in ancestors:
                ancestor.find_named_blocks(loader, named_blocks)
            writer = _CodeWriter(buffer, named_blocks, loader,
                                 ancestors[0].template)
            ancestors[0].generate(writer)
            return buffer.getvalue()
        finally:
            buffer.close()

    def _get_ancestors(self, loader):
        ancestors = [self.file]
        for chunk in self.file.body.chunks:
            if isinstance(chunk, _ExtendsBlock):
                if not loader:
                    raise ParseError("{% extends %} block found, but no "
                                     "template loader")
                template = loader.load(chunk.name, self.name)
                ancestors.extend(template._get_ancestors(loader))
        return ancestors


class BaseLoader(object):
    """Base class for template loaders.

    You must use a template loader to use template constructs like
    ``{% extends %}`` and ``{% include %}``. The loader caches all
    templates after they are loaded the first time.
    """
    def __init__(self, autoescape=_DEFAULT_AUTOESCAPE, namespace=None,
                 whitespace=None):
        """Construct a template loader.

        :arg str autoescape: The name of a function in the template
            namespace, such as "xhtml_escape", or ``None`` to disable
            autoescaping by default.
        :arg dict namespace: A dictionary to be added to the default template
            namespace, or ``None``.
        :arg str whitespace: A string specifying default behavior for
            whitespace in templates; see `filter_whitespace` for options.
            Default is "single" for files ending in ".html" and ".js" and
            "all" for other files.

        .. versionchanged:: 4.3
           Added ``whitespace`` parameter.
        """
        self.autoescape = autoescape
        self.namespace = namespace or {}
        self.whitespace = whitespace
        self.templates = {}
        # self.lock protects self.templates.  It's a reentrant lock
        # because templates may load other templates via `include` or
        # `extends`.  Note that thanks to the GIL this code would be safe
        # even without the lock, but could lead to wasted work as multiple
        # threads tried to compile the same template simultaneously.
        self.lock = threading.RLock()

    def reset(self):
        """Resets the cache of compiled templates."""
        with self.lock:
            self.templates = {}

    def resolve_path(self, name, parent_path=None):
        """Converts a possibly-relative path to absolute (used internally)."""
        raise NotImplementedError()

    def load(self, name, parent_path=None):
        """Loads a template."""
        name = self.resolve_path(name, parent_path=parent_path)
        with self.lock:
            if name not in self.templates:
                self.templates[name] = self._create_template(name)
            return self.templates[name]

    def _create_template(self, name):
        raise NotImplementedError()


class Loader(BaseLoader):
    """A template loader that loads from a single root directory.
    """
    def __init__(self, root_directory, **kwargs):
        super(Loader, self).__init__(**kwargs)
        self.root = os.path.abspath(root_directory)

    def resolve_path(self, name, parent_path=None):
        if parent_path and not parent_path.startswith("<") and \
            not parent_path.startswith("/") and \
                not name.startswith("/"):
            current_path = os.path.join(self.root, parent_path)
            file_dir = os.path.dirname(os.path.abspath(current_path))
            relative_path = os.path.abspath(os.path.join(file_dir, name))
            if relative_path.startswith(self.root):
                name = relative_path[len(self.root) + 1:]
        return name

    def _create_template(self, name):
        path = os.path.join(self.root, name)
        with open(path, "rb") as f:
            template = Template(f.read(), name=name, loader=self)
            return template


class DictLoader(BaseLoader):
    """A template loader that loads from a dictionary."""
    def __init__(self, dict, **kwargs):
        super(DictLoader, self).__init__(**kwargs)
        self.dict = dict

    def resolve_path(self, name, parent_path=None):
        if parent_path and not parent_path.startswith("<") and \
            not parent_path.startswith("/") and \
                not name.startswith("/"):
            file_dir = posixpath.dirname(parent_path)
            name = posixpath.normpath(posixpath.join(file_dir, name))
        return name

    def _create_template(self, name):
        return Template(self.dict[name], name=name, loader=self)


class _Node(object):
    def each_child(self):
        return ()

    def generate(self, writer):
        raise NotImplementedError()

    def find_named_blocks(self, loader, named_blocks):
        for child in self.each_child():
            child.find_named_blocks(loader, named_blocks)


class _File(_Node):
    def __init__(self, template, body):
        self.template = template
        self.body = body
        self.line = 0

    def generate(self, writer):
        writer.write_line("def _tt_execute():", self.line)
        with writer.indent():
            writer.write_line("_tt_buffer = []", self.line)
            writer.write_line("_tt_append = _tt_buffer.append", self.line)
            self.body.generate(writer)
            writer.write_line("return _tt_utf8('').join(_tt_buffer)", self.line)

    def each_child(self):
        return (self.body,)


class _ChunkList(_Node):
    def __init__(self, chunks):
        self.chunks = chunks

    def generate(self, writer):
        for chunk in self.chunks:
            chunk.generate(writer)

    def each_child(self):
        return self.chunks


class _NamedBlock(_Node):
    def __init__(self, name, body, template, line):
        self.name = name
        self.body = body
        self.template = template
        self.line = line

    def each_child(self):
        return (self.body,)

    def generate(self, writer):
        block = writer.named_blocks[self.name]
        with writer.include(block.template, self.line):
            block.body.generate(writer)

    def find_named_blocks(self, loader, named_blocks):
        named_blocks[self.name] = self
        _Node.find_named_blocks(self, loader, named_blocks)


class _ExtendsBlock(_Node):
    def __init__(self, name):
        self.name = name


class _IncludeBlock(_Node):
    def __init__(self, name, reader, line):
        self.name = name
        self.template_name = reader.name
        self.line = line

    def find_named_blocks(self, loader, named_blocks):
        included = loader.load(self.name, self.template_name)
        included.file.find_named_blocks(loader, named_blocks)

    def generate(self, writer):
        included = writer.loader.load(self.name, self.template_name)
        with writer.include(included, self.line):
            included.file.body.generate(writer)


class _ApplyBlock(_Node):
    def __init__(self, method, line, body=None):
        self.method = method
        self.line = line
        self.body = body

    def each_child(self):
        return (self.body,)

    def generate(self, writer):
        method_name = "_tt_apply%d" % writer.apply_counter
        writer.apply_counter += 1
        writer.write_line("def %s():" % method_name, self.line)
        with writer.indent():
            writer.write_line("_tt_buffer = []", self.line)
            writer.write_line("_tt_append = _tt_buffer.append", self.line)
            self.body.generate(writer)
            writer.write_line("return _tt_utf8('').join(_tt_buffer)", self.line)
        writer.write_line("_tt_append(_tt_utf8(%s(%s())))" % (
            self.method, method_name), self.line)


class _ControlBlock(_Node):
    def __init__(self, statement, line, body=None):
        self.statement = statement
        self.line = line
        self.body = body

    def each_child(self):
        return (self.body,)

    def generate(self, writer):
        writer.write_line("%s:" % self.statement, self.line)
        with writer.indent():
            self.body.generate(writer)
            # Just in case the body was empty
            writer.write_line("pass", self.line)


class _IntermediateControlBlock(_Node):
    def __init__(self, statement, line):
        self.statement = statement
        self.line = line

    def generate(self, writer):
        # In case the previous block was empty
        writer.write_line("pass", self.line)
        writer.write_line("%s:" % self.statement, self.line, writer.indent_size() - 1)


class _Statement(_Node):
    def __init__(self, statement, line):
        self.statement = statement
        self.line = line

    def generate(self, writer):
        writer.write_line(self.statement, self.line)


class _Expression(_Node):
    def __init__(self, expression, line, raw=False):
        self.expression = expression
        self.line = line
        self.raw = raw

    def generate(self, writer):
        writer.write_line("_tt_tmp = %s" % self.expression, self.line)
        writer.write_line("if isinstance(_tt_tmp, _tt_string_types):"
                          " _tt_tmp = _tt_utf8(_tt_tmp)", self.line)
        writer.write_line("else: _tt_tmp = _tt_utf8(str(_tt_tmp))", self.line)
        if not self.raw and writer.current_template.autoescape is not None:
            # In python3 functions like xhtml_escape return unicode,
            # so we have to convert to utf8 again.
            writer.write_line("_tt_tmp = _tt_utf8(%s(_tt_tmp))" %
                              writer.current_template.autoescape, self.line)
        writer.write_line("_tt_append(_tt_tmp)", self.line)


class _Module(_Expression):
    def __init__(self, expression, line):
        super(_Module, self).__init__("_tt_modules." + expression, line,
                                      raw=True)


class _Text(_Node):
    def __init__(self, value, line, whitespace):
        self.value = value
        self.line = line
        self.whitespace = whitespace

    def generate(self, writer):
        value = self.value

        # Compress whitespace if requested, with a crude heuristic to avoid
        # altering preformatted whitespace.
        if "<pre>" not in value:
            value = filter_whitespace(self.whitespace, value)

        if value:
            writer.write_line('_tt_append(%r)' % escape.utf8(value), self.line)


class ParseError(Exception):
    """Raised for template syntax errors.

    ``ParseError`` instances have ``filename`` and ``lineno`` attributes
    indicating the position of the error.

    .. versionchanged:: 4.3
       Added ``filename`` and ``lineno`` attributes.
    """
    def __init__(self, message, filename=None, lineno=0):
        self.message = message
        # The names "filename" and "lineno" are chosen for consistency
        # with python SyntaxError.
        self.filename = filename
        self.lineno = lineno

    def __str__(self):
        return '%s at %s:%d' % (self.message, self.filename, self.lineno)


class _CodeWriter(object):
    def __init__(self, file, named_blocks, loader, current_template):
        self.file = file
        self.named_blocks = named_blocks
        self.loader = loader
        self.current_template = current_template
        self.apply_counter = 0
        self.include_stack = []
        self._indent = 0

    def indent_size(self):
        return self._indent

    def indent(self):
        class Indenter(object):
            def __enter__(_):
                self._indent += 1
                return self

            def __exit__(_, *args):
                assert self._indent > 0
                self._indent -= 1

        return Indenter()

    def include(self, template, line):
        self.include_stack.append((self.current_template, line))
        self.current_template = template

        class IncludeTemplate(object):
            def __enter__(_):
                return self

            def __exit__(_, *args):
                self.current_template = self.include_stack.pop()[0]

        return IncludeTemplate()

    def write_line(self, line, line_number, indent=None):
        if indent is None:
            indent = self._indent
        line_comment = '  # %s:%d' % (self.current_template.name, line_number)
        if self.include_stack:
            ancestors = ["%s:%d" % (tmpl.name, lineno)
                         for (tmpl, lineno) in self.include_stack]
            line_comment += ' (via %s)' % ', '.join(reversed(ancestors))
        print("    " * indent + line + line_comment, file=self.file)


class _TemplateReader(object):
    def __init__(self, name, text, whitespace):
        self.name = name
        self.text = text
        self.whitespace = whitespace
        self.line = 1
        self.pos = 0

    def find(self, needle, start=0, end=None):
        assert start >= 0, start
        pos = self.pos
        start += pos
        if end is None:
            index = self.text.find(needle, start)
        else:
            end += pos
            assert end >= start
            index = self.text.find(needle, start, end)
        if index != -1:
            index -= pos
        return index

    def consume(self, count=None):
        if count is None:
            count = len(self.text) - self.pos
        newpos = self.pos + count
        self.line += self.text.count("\n", self.pos, newpos)
        s = self.text[self.pos:newpos]
        self.pos = newpos
        return s

    def remaining(self):
        return len(self.text) - self.pos

    def __len__(self):
        return self.remaining()

    def __getitem__(self, key):
        if type(key) is slice:
            size = len(self)
            start, stop, step = key.indices(size)
            if start is None:
                start = self.pos
            else:
                start += self.pos
            if stop is not None:
                stop += self.pos
            return self.text[slice(start, stop, step)]
        elif key < 0:
            return self.text[key]
        else:
            return self.text[self.pos + key]

    def __str__(self):
        return self.text[self.pos:]

    def raise_parse_error(self, msg):
        raise ParseError(msg, self.name, self.line)


def _format_code(code):
    lines = code.splitlines()
    format = "%%%dd  %%s\n" % len(repr(len(lines) + 1))
    return "".join([format % (i + 1, line) for (i, line) in enumerate(lines)])


def _parse(reader, template, in_block=None, in_loop=None):
    body = _ChunkList([])
    while True:
        # Find next template directive
        curly = 0
        while True:
            curly = reader.find("{", curly)
            if curly == -1 or curly + 1 == reader.remaining():
                # EOF
                if in_block:
                    reader.raise_parse_error(
                        "Missing {%% end %%} block for %s" % in_block)
                body.chunks.append(_Text(reader.consume(), reader.line,
                                         reader.whitespace))
                return body
            # If the first curly brace is not the start of a special token,
            # start searching from the character after it
            if reader[curly + 1] not in ("{", "%", "#"):
                curly += 1
                continue
            # When there are more than 2 curlies in a row, use the
            # innermost ones.  This is useful when generating languages
            # like latex where curlies are also meaningful
            if (curly + 2 < reader.remaining() and
                    reader[curly + 1] == '{' and reader[curly + 2] == '{'):
                curly += 1
                continue
            break

        # Append any text before the special token
        if curly > 0:
            cons = reader.consume(curly)
            body.chunks.append(_Text(cons, reader.line,
                                     reader.whitespace))

        start_brace = reader.consume(2)
        line = reader.line

        # Template directives may be escaped as "{{!" or "{%!".
        # In this case output the braces and consume the "!".
        # This is especially useful in conjunction with jquery templates,
        # which also use double braces.
        if reader.remaining() and reader[0] == "!":
            reader.consume(1)
            body.chunks.append(_Text(start_brace, line,
                                     reader.whitespace))
            continue

        # Comment
        if start_brace == "{#":
            end = reader.find("#}")
            if end == -1:
                reader.raise_parse_error("Missing end comment #}")
            contents = reader.consume(end).strip()
            reader.consume(2)
            continue

        # Expression
        if start_brace == "{{":
            end = reader.find("}}")
            if end == -1:
                reader.raise_parse_error("Missing end expression }}")
            contents = reader.consume(end).strip()
            reader.consume(2)
            if not contents:
                reader.raise_parse_error("Empty expression")
            body.chunks.append(_Expression(contents, line))
            continue

        # Block
        assert start_brace == "{%", start_brace
        end = reader.find("%}")
        if end == -1:
            reader.raise_parse_error("Missing end block %}")
        contents = reader.consume(end).strip()
        reader.consume(2)
        if not contents:
            reader.raise_parse_error("Empty block tag ({% %})")

        operator, space, suffix = contents.partition(" ")
        suffix = suffix.strip()

        # Intermediate ("else", "elif", etc) blocks
        intermediate_blocks = {
            "else": set(["if", "for", "while", "try"]),
            "elif": set(["if"]),
            "except": set(["try"]),
            "finally": set(["try"]),
        }
        allowed_parents = intermediate_blocks.get(operator)
        if allowed_parents is not None:
            if not in_block:
                reader.raise_parse_error("%s outside %s block" %
                                         (operator, allowed_parents))
            if in_block not in allowed_parents:
                reader.raise_parse_error(
                    "%s block cannot be attached to %s block" %
                    (operator, in_block))
            body.chunks.append(_IntermediateControlBlock(contents, line))
            continue

        # End tag
        elif operator == "end":
            if not in_block:
                reader.raise_parse_error("Extra {% end %} block")
            return body

        elif operator in ("extends", "include", "set", "import", "from",
                          "comment", "autoescape", "whitespace", "raw",
                          "module"):
            if operator == "comment":
                continue
            if operator == "extends":
                suffix = suffix.strip('"').strip("'")
                if not suffix:
                    reader.raise_parse_error("extends missing file path")
                block = _ExtendsBlock(suffix)
            elif operator in ("import", "from"):
                if not suffix:
                    reader.raise_parse_error("import missing statement")
                block = _Statement(contents, line)
            elif operator == "include":
                suffix = suffix.strip('"').strip("'")
                if not suffix:
                    reader.raise_parse_error("include missing file path")
                block = _IncludeBlock(suffix, reader, line)
            elif operator == "set":
                if not suffix:
                    reader.raise_parse_error("set missing statement")
                block = _Statement(suffix, line)
            elif operator == "autoescape":
                fn = suffix.strip()
                if fn == "None":
                    fn = None
                template.autoescape = fn
                continue
            elif operator == "whitespace":
                mode = suffix.strip()
                # Validate the selected mode
                filter_whitespace(mode, '')
                reader.whitespace = mode
                continue
            elif operator == "raw":
                block = _Expression(suffix, line, raw=True)
            elif operator == "module":
                block = _Module(suffix, line)
            body.chunks.append(block)
            continue

        elif operator in ("apply", "block", "try", "if", "for", "while"):
            # parse inner body recursively
            if operator in ("for", "while"):
                block_body = _parse(reader, template, operator, operator)
            elif operator == "apply":
                # apply creates a nested function so syntactically it's not
                # in the loop.
                block_body = _parse(reader, template, operator, None)
            else:
                block_body = _parse(reader, template, operator, in_loop)

            if operator == "apply":
                if not suffix:
                    reader.raise_parse_error("apply missing method name")
                block = _ApplyBlock(suffix, line, block_body)
            elif operator == "block":
                if not suffix:
                    reader.raise_parse_error("block missing name")
                block = _NamedBlock(suffix, block_body, template, line)
            else:
                block = _ControlBlock(contents, line, block_body)
            body.chunks.append(block)
            continue

        elif operator in ("break", "continue"):
            if not in_loop:
                reader.raise_parse_error("%s outside %s block" %
                                         (operator, set(["for", "while"])))
            body.chunks.append(_Statement(contents, line))
            continue

        else:
            reader.raise_parse_error("unknown operator: %r" % operator)



import shutil

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(autouse=True)
def thin_dir(salt_ssh_cli):
    try:
        yield
    finally:
        ret = salt_ssh_cli.run("config.get", "thin_dir")
        assert ret.returncode == 0
        thin_dir_path = ret.data
        shutil.rmtree(thin_dir_path, ignore_errors=True)


def test_ssh_mine_get(salt_ssh_cli):
    """
    test salt-ssh with mine
    """
    ret = salt_ssh_cli.run("mine.get", "localhost", "test.arg")
    assert ret.returncode == 0
    assert ret.data
    assert "localhost" in ret.data
    assert "args" in ret.data["localhost"]
    assert ret.data["localhost"]["args"] == ["itworked"]



"""
The cp module is used to execute the logic used by the salt-cp command
line application, salt-cp is NOT intended to broadcast large files, it is
intended to handle text files.
Salt-cp can be used to distribute configuration files
"""

import base64
import errno
import logging
import os
import re
import sys

import salt.client
import salt.output
import salt.utils.files
import salt.utils.gzip_util
import salt.utils.itertools
import salt.utils.minions
import salt.utils.parsers
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.verify

log = logging.getLogger(__name__)


class SaltCPCli(salt.utils.parsers.SaltCPOptionParser):
    """
    Run the salt-cp command line client
    """

    def run(self):
        """
        Execute salt-cp
        """
        self.parse_args()
        cp_ = SaltCP(self.config)
        cp_.run()


class SaltCP:
    """
    Create a salt cp object, used to distribute simple files with salt
    """

    def __init__(self, opts):
        self.opts = opts
        self.is_windows = salt.utils.platform.is_windows()

    def _mode(self, path):
        if self.is_windows:
            return None
        try:
            return int(oct(os.stat(path).st_mode)[-4:], 8)
        except (TypeError, IndexError, ValueError):
            return None

    def _recurse(self, path):
        """
        Get a list of all specified files
        """
        files = {}
        empty_dirs = []
        try:
            sub_paths = os.listdir(path)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                # Path does not exist
                sys.stderr.write(f"{path} does not exist\n")
                sys.exit(42)
            elif exc.errno in (errno.EINVAL, errno.ENOTDIR):
                # Path is a file (EINVAL on Windows, ENOTDIR otherwise)
                files[path] = self._mode(path)
        else:
            if not sub_paths:
                empty_dirs.append(path)
            for fn_ in sub_paths:
                files_, empty_dirs_ = self._recurse(os.path.join(path, fn_))
                files.update(files_)
                empty_dirs.extend(empty_dirs_)

        return files, empty_dirs

    def _list_files(self):
        files = {}
        empty_dirs = set()
        for fn_ in self.opts["src"]:
            files_, empty_dirs_ = self._recurse(fn_)
            files.update(files_)
            empty_dirs.update(empty_dirs_)
        return files, sorted(empty_dirs)

    def _file_dict(self, fn_):
        """
        Take a path and return the contents of the file as a string
        """
        if not os.path.isfile(fn_):
            err = f"The referenced file, {fn_} is not available."
            sys.stderr.write(err + "\n")
            sys.exit(42)
        with salt.utils.files.fopen(fn_, "r") as fp_:
            data = fp_.read()
        return {fn_: data}

    def _load_files(self):
        """
        Parse the files indicated in opts['src'] and load them into a python
        object for transport
        """
        files = {}
        for fn_ in self.opts["src"]:
            if os.path.isfile(fn_):
                files.update(self._file_dict(fn_))
            elif os.path.isdir(fn_):
                salt.utils.stringutils.print_cli(
                    "{} is a directory, only files are supported "
                    'in non-chunked mode. Use "--chunked" command '
                    "line argument.".format(fn_)
                )
                sys.exit(1)
        return files

    def run(self):
        """
        Make the salt client call
        """
        if self.opts["chunked"]:
            ret = self.run_chunked()
        else:
            ret = self.run_oldstyle()

        salt.output.display_output(ret, self.opts.get("output", "nested"), self.opts)

    def run_oldstyle(self):
        """
        Make the salt client call in old-style all-in-one call method
        """
        arg = [self._load_files(), self.opts["dest"]]
        args = [
            self.opts["tgt"],
            "cp.recv",
            arg,
            self.opts["timeout"],
        ]

        selected_target_option = self.opts.get("selected_target_option", None)
        if selected_target_option is not None:
            args.append(selected_target_option)

        with salt.client.get_local_client(self.opts["conf_file"]) as local:
            return local.cmd(*args)

    def run_chunked(self):
        """
        Make the salt client call in the new fasion chunked multi-call way
        """
        files, empty_dirs = self._list_files()
        dest = self.opts["dest"]
        gzip = self.opts["gzip"]
        tgt = self.opts["tgt"]
        timeout = self.opts["timeout"]
        selected_target_option = self.opts.get("selected_target_option")

        dest_is_dir = (
            bool(empty_dirs) or len(files) > 1 or bool(re.search(r"[\\/]$", dest))
        )

        reader = (
            salt.utils.gzip_util.compress_file
            if gzip
            else salt.utils.itertools.read_file
        )

        _res = salt.utils.minions.CkMinions(self.opts).check_minions(
            tgt, tgt_type=selected_target_option or "glob"
        )
        minions = _res["minions"]

        def _get_remote_path(fn_):
            if fn_ in self.opts["src"]:
                # This was a filename explicitly passed on the CLI
                return (
                    os.path.join(dest, os.path.basename(fn_)) if dest_is_dir else dest
                )
            else:
                for path in self.opts["src"]:
                    relpath = os.path.relpath(fn_, path + os.sep)
                    if relpath.startswith(parent):
                        # File is not within this dir
                        continue
                    return os.path.join(dest, os.path.basename(path), relpath)
                else:  # pylint: disable=useless-else-on-loop
                    # Should not happen
                    log.error("Failed to find remote path for %s", fn_)
                    return None

        ret = {}
        parent = ".." + os.sep

        with salt.client.get_local_client(self.opts["conf_file"]) as local:
            for fn_, mode in files.items():
                remote_path = _get_remote_path(fn_)

                index = 1
                failed = {}
                for chunk in reader(fn_, chunk_size=self.opts["salt_cp_chunk_size"]):
                    chunk = base64.b64encode(salt.utils.stringutils.to_bytes(chunk))
                    append = index > 1
                    log.debug(
                        "Copying %s to %starget '%s' as %s%s",
                        fn_,
                        f"{selected_target_option} " if selected_target_option else "",
                        tgt,
                        remote_path,
                        f" (chunk #{index})" if append else "",
                    )
                    args = [
                        tgt,
                        "cp.recv_chunked",
                        [remote_path, chunk, append, gzip, mode],
                        timeout,
                    ]
                    if selected_target_option is not None:
                        args.append(selected_target_option)

                    result = local.cmd(*args)

                    if not result:
                        # Publish failed
                        msg = (
                            "Publish failed.{} It may be necessary to "
                            "decrease salt_cp_chunk_size (current value: "
                            "{})".format(
                                " File partially transferred." if index > 1 else "",
                                self.opts["salt_cp_chunk_size"],
                            )
                        )
                        for minion in minions:
                            ret.setdefault(minion, {})[remote_path] = msg
                        break

                    for minion_id, minion_ret in result.items():
                        ret.setdefault(minion_id, {})[remote_path] = minion_ret
                        # Catch first error message for a given minion, we will
                        # rewrite the results after we're done iterating through
                        # the chunks.
                        if minion_ret is not True and minion_id not in failed:
                            failed[minion_id] = minion_ret

                    index += 1

                for minion_id, msg in failed.items():
                    ret[minion_id][remote_path] = msg

            for dirname in empty_dirs:
                remote_path = _get_remote_path(dirname)
                log.debug(
                    "Creating empty dir %s on %starget '%s'",
                    dirname,
                    f"{selected_target_option} " if selected_target_option else "",
                    tgt,
                )
                args = [tgt, "cp.recv_chunked", [remote_path, None], timeout]
                if selected_target_option is not None:
                    args.append(selected_target_option)

                for minion_id, minion_ret in local.cmd(*args).items():
                    ret.setdefault(minion_id, {})[remote_path] = minion_ret

            return ret



import logging
import shutil
import types

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def minion_config_overrides(tmp_path_factory):
    temp_path = tmp_path_factory.mktemp("aliases-temp")
    aliases_file = temp_path / "aliases-file"
    try:
        yield {
            "aliases.file": str(aliases_file),
            "integration.test": True,
        }
    finally:
        shutil.rmtree(str(temp_path), ignore_errors=True)


@pytest.fixture(scope="module")
def aliases(modules):
    return modules.aliases


@pytest.fixture
def alias(aliases):
    ret = aliases.set_target(alias="fred", target="bob")
    assert ret is True
    return types.SimpleNamespace(name="fred", target="bob")


def test_set_target(aliases, alias):
    """
    aliases.set_target and aliases.get_target
    """
    ret = aliases.get_target(alias.name)
    assert ret == alias.target


def test_has_target(aliases, alias):
    """
    aliases.set_target and aliases.has_target
    """
    ret = aliases.has_target(alias.name, alias.target)
    assert ret is True


def test_list_aliases(aliases, alias):
    """
    aliases.list_aliases
    """
    ret = aliases.list_aliases()
    assert isinstance(ret, dict)
    assert alias.name in ret


def test_rm_alias(aliases, alias):
    """
    aliases.rm_alias
    """
    ret = aliases.rm_alias(alias=alias.name)
    assert ret is True
    ret = aliases.list_aliases()
    assert isinstance(ret, dict)
    assert ret == {}



import subprocess

import pytest

pytestmark = [
    pytest.mark.skip_unless_on_linux,
]


@pytest.mark.usefixtures("salt_minion")
def test_system_config(grains):
    """
    Test system config
    """
    if grains["os_family"] == "RedHat":
        if grains["osfinger"] in (
            "AlmaLinux-8",
            "AlmaLinux-9",
            "Rocky Linux-8",
            "Rocky Linux-9",
            "CentOS Stream-8",
            "CentOS Linux-8",
            "CentOS Stream-9",
            "Fedora Linux-40",
            "VMware Photon OS-4",
            "VMware Photon OS-5",
            "Amazon Linux-2023",
        ):
            expected_retcode = 0
        else:
            expected_retcode = 1
        ret = subprocess.call(
            "systemctl show -p ${config} salt-minion.service", shell=True
        )
        assert ret == expected_retcode

    elif grains["os_family"] == "Debian":
        if grains["osfinger"] == "Debian-9":
            expected_retcode = 1
        else:
            expected_retcode = 0
        ret = subprocess.call(
            "systemctl show -p ${config} salt-minion.service", shell=True
        )
        assert ret == expected_retcode



"""
Monitor Server with Server Density
==================================

.. versionadded:: 2014.7.0

`Server Density <https://www.serverdensity.com/>`_
Is a hosted monitoring service.

.. warning::

    This state module is beta. It might be changed later to include more or
    less automation.

.. note::

    This state module requires a pillar for authentication with Server Density
    To install a v1 agent:

    .. code-block:: yaml

        serverdensity:
          api_token: "b97da80a41c4f61bff05975ee51eb1aa"
          account_url: "https://your-account.serverdensity.io"

    To install a v2 agent:

    .. code-block:: yaml

        serverdensity:
          api_token: "b97da80a41c4f61bff05975ee51eb1aa"
          account_name: "your-account"

.. note::

    Although Server Density allows duplicate device names in its database, this
    module will raise an exception if you try monitoring devices with the same
    name.


Example:

.. code-block:: yaml

    'server_name':
      serverdensity_device.monitored
"""

import logging

import salt.utils.json

# TODO:
#
#  Add a plugin support
#  Add notification support


log = logging.getLogger(__name__)


def _get_salt_params():
    """
    Try to get all sort of parameters for Server Density server info.

    NOTE: Missing publicDNS and publicIPs parameters. There might be way of
    getting them with salt-cloud.
    """
    all_stats = __salt__["status.all_status"]()
    all_grains = __salt__["grains.items"]()
    params = {}
    try:
        params["name"] = all_grains["id"]
        params["hostname"] = all_grains["host"]
        if all_grains["kernel"] == "Darwin":
            sd_os = {"code": "mac", "name": "Mac"}
        else:
            sd_os = {"code": all_grains["kernel"].lower(), "name": all_grains["kernel"]}
        params["os"] = salt.utils.json.dumps(sd_os)
        params["cpuCores"] = all_stats["cpuinfo"]["cpu cores"]
        params["installedRAM"] = str(
            int(all_stats["meminfo"]["MemTotal"]["value"]) / 1024
        )
        params["swapSpace"] = str(
            int(all_stats["meminfo"]["SwapTotal"]["value"]) / 1024
        )
        params["privateIPs"] = salt.utils.json.dumps(all_grains["fqdn_ip4"])
        params["privateDNS"] = salt.utils.json.dumps(all_grains["fqdn"])
    except KeyError:
        pass

    return params


def monitored(
    name, group=None, salt_name=True, salt_params=True, agent_version=1, **params
):
    """
    Device is monitored with Server Density.

    name
        Device name in Server Density.

    salt_name
        If ``True`` (default), takes the name from the ``id`` grain. If
        ``False``, the provided name is used.

    group
        Group name under with device will appear in Server Density dashboard.
        Default - `None`.

    agent_version
        The agent version you want to use. Valid values are 1 or 2.
        Default - 1.

    salt_params
        If ``True`` (default), needed config parameters will be sourced from
        grains and from :mod:`status.all_status
        <salt.modules.status.all_status>`.

    params
        Add parameters that you want to appear in the Server Density dashboard.
        Will overwrite the `salt_params` parameters. For more info, see the
        `API docs`__.

    .. __: https://apidocs.serverdensity.com/Inventory/Devices/Creating

    Usage example:

    .. code-block:: yaml

        'server_name':
          serverdensity_device.monitored

    .. code-block:: yaml

        'server_name':
          serverdensity_device.monitored:
            - group: web-servers

    .. code-block:: yaml

        'my_special_server':
          serverdensity_device.monitored:
            - salt_name: False
            - group: web-servers
            - cpuCores: 2
            - os: '{"code": "linux", "name": "Linux"}'
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    params_from_salt = _get_salt_params()

    if salt_name:
        name = params_from_salt.pop("name")
        ret["name"] = name
    else:
        params_from_salt.pop("name")

    if group:
        params["group"] = group

    if agent_version != 2:
        # Anything different from 2 will fallback into the v1.
        agent_version = 1

    # override salt_params with given params
    if salt_params:
        for key, value in params.items():
            params_from_salt[key] = value
        params_to_use = params_from_salt
    else:
        params_to_use = params

    device_in_sd = True if __salt__["serverdensity_device.ls"](name=name) else False
    sd_agent_installed = True if "sd-agent" in __salt__["pkg.list_pkgs"]() else False

    if device_in_sd and sd_agent_installed:
        ret["result"] = True
        ret["comment"] = (
            "Such server name already exists in this Server Density account. And"
            " sd-agent is installed"
        )
        ret["changes"] = {}
        return ret

    if not device_in_sd:
        device = __salt__["serverdensity_device.create"](name, **params_from_salt)
        agent_key = device["agentKey"]
        ret["comment"] = "Device created in Server Density db."
        ret["changes"] = {"device_created": device}
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Device set to be created in Server Density db."
            return ret
    elif device_in_sd:
        device = __salt__["serverdensity_device.ls"](name=name)[0]
        agent_key = device["agentKey"]
        ret["comment"] = "Device was already in Server Density db."
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to create device in Server Density DB and this device does not"
            " exist in db either."
        )
        ret["changes"] = {}
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                "Agent is not installed and device is not in the Server Density DB"
            )
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = (
            "Server Density agent is set to be installed and device created in the"
            " Server Density DB"
        )
        return ret

    installed_agent = __salt__["serverdensity_device.install_agent"](
        agent_key, agent_version
    )

    ret["result"] = True
    ret["comment"] = (
        "Successfully installed agent and created device in Server Density db."
    )
    ret["changes"] = {"created_device": device, "installed_agent": installed_agent}
    return ret



"""
Related to zypp_plugins_test.py module.
"""


class Plugin:
    """
    Bogus module for Zypp Plugins tests.
    """

    def ack(self):
        """
        Acknowledge that the plugin had finished the transaction
        Returns:

        """

    def main(self):
        """
        Register plugin
        Returns:

        """


class BogusIO:
    """
    Read/write logger.
    """

    def __init__(self):
        self.content = list()
        self.closed = False

    def __str__(self):
        return "\n".join(self.content)

    def __call__(self, *args, **kwargs):
        self.path, self.mode = args
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __enter__(self):
        return self

    def write(self, data):
        """
        Simulate writing data
        Args:
            data:

        Returns:

        """
        self.content.append(data)

    def close(self):
        """
        Simulate closing the IO object.
        Returns:

        """
        self.closed = True



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

import time

import pytest

import salt.master
import salt.utils.platform
from salt import auth
from salt.exceptions import SaltDeserializationError
from tests.support.case import ModuleCase
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase


class LoadAuthTestCase(TestCase):
    def setUp(self):  # pylint: disable=W0221
        patches = (
            ("salt.payload.Serial", None),
            (
                "salt.loader.auth",
                dict(
                    return_value={
                        "pam.auth": "fake_func_str",
                        "pam.groups": "fake_groups_function_str",
                    }
                ),
            ),
            (
                "salt.loader.eauth_tokens",
                dict(
                    return_value={
                        "localfs.mk_token": "fake_func_mktok",
                        "localfs.get_token": "fake_func_gettok",
                        "localfs.rm_roken": "fake_func_rmtok",
                    }
                ),
            ),
        )
        for mod, mock in patches:
            if mock:
                patcher = patch(mod, **mock)
            else:
                patcher = patch(mod)
            patcher.start()
            self.addCleanup(patcher.stop)
        self.lauth = auth.LoadAuth({})  # Load with empty opts

    def test_get_tok_with_broken_file_will_remove_bad_token(self):
        fake_get_token = MagicMock(side_effect=SaltDeserializationError("hi"))
        patch_opts = patch.dict(self.lauth.opts, {"eauth_tokens": "testfs"})
        patch_get_token = patch.dict(
            self.lauth.tokens,
            {"testfs.get_token": fake_get_token},
        )
        mock_rm_token = MagicMock()
        patch_rm_token = patch.object(self.lauth, "rm_token", mock_rm_token)
        with patch_opts, patch_get_token, patch_rm_token:
            expected_token = "fnord"
            self.lauth.get_tok(expected_token)
            mock_rm_token.assert_called_with(expected_token)

    def test_get_tok_with_no_expiration_should_remove_bad_token(self):
        fake_get_token = MagicMock(return_value={"no_expire_here": "Nope"})
        patch_opts = patch.dict(self.lauth.opts, {"eauth_tokens": "testfs"})
        patch_get_token = patch.dict(
            self.lauth.tokens,
            {"testfs.get_token": fake_get_token},
        )
        mock_rm_token = MagicMock()
        patch_rm_token = patch.object(self.lauth, "rm_token", mock_rm_token)
        with patch_opts, patch_get_token, patch_rm_token:
            expected_token = "fnord"
            self.lauth.get_tok(expected_token)
            mock_rm_token.assert_called_with(expected_token)

    def test_get_tok_with_expire_before_current_time_should_remove_token(self):
        fake_get_token = MagicMock(return_value={"expire": time.time() - 1})
        patch_opts = patch.dict(self.lauth.opts, {"eauth_tokens": "testfs"})
        patch_get_token = patch.dict(
            self.lauth.tokens,
            {"testfs.get_token": fake_get_token},
        )
        mock_rm_token = MagicMock()
        patch_rm_token = patch.object(self.lauth, "rm_token", mock_rm_token)
        with patch_opts, patch_get_token, patch_rm_token:
            expected_token = "fnord"
            self.lauth.get_tok(expected_token)
            mock_rm_token.assert_called_with(expected_token)

    def test_get_tok_with_valid_expiration_should_return_token(self):
        expected_token = {"expire": time.time() + 1}
        fake_get_token = MagicMock(return_value=expected_token)
        patch_opts = patch.dict(self.lauth.opts, {"eauth_tokens": "testfs"})
        patch_get_token = patch.dict(
            self.lauth.tokens,
            {"testfs.get_token": fake_get_token},
        )
        mock_rm_token = MagicMock()
        patch_rm_token = patch.object(self.lauth, "rm_token", mock_rm_token)
        with patch_opts, patch_get_token, patch_rm_token:
            token_name = "fnord"
            actual_token = self.lauth.get_tok(token_name)
            mock_rm_token.assert_not_called()
            assert expected_token is actual_token, "Token was not returned"

    def test_load_name(self):
        valid_eauth_load = {
            "username": "test_user",
            "show_timeout": False,
            "test_password": "",
            "eauth": "pam",
        }

        # Test a case where the loader auth doesn't have the auth type
        without_auth_type = dict(valid_eauth_load)
        without_auth_type.pop("eauth")
        ret = self.lauth.load_name(without_auth_type)
        self.assertEqual(
            ret, "", "Did not bail when the auth loader didn't have the auth type."
        )

        # Test a case with valid params
        with patch(
            "salt.utils.args.arg_lookup",
            MagicMock(return_value={"args": ["username", "password"]}),
        ) as format_call_mock:
            expected_ret = call("fake_func_str")
            ret = self.lauth.load_name(valid_eauth_load)
            format_call_mock.assert_has_calls((expected_ret,), any_order=True)
            self.assertEqual(ret, "test_user")

    def test_get_groups(self):
        valid_eauth_load = {
            "username": "test_user",
            "show_timeout": False,
            "test_password": "",
            "eauth": "pam",
        }
        with patch("salt.utils.args.format_call") as format_call_mock:
            expected_ret = call(
                "fake_groups_function_str",
                {
                    "username": "test_user",
                    "test_password": "",
                    "show_timeout": False,
                    "eauth": "pam",
                },
                expected_extra_kws=auth.AUTH_INTERNAL_KEYWORDS,
            )
            self.lauth.get_groups(valid_eauth_load)
            format_call_mock.assert_has_calls((expected_ret,), any_order=True)


class MasterACLTestCase(ModuleCase):
    """
    A class to check various aspects of the publisher ACL system
    """

    def setUp(self):
        self.fire_event_mock = MagicMock(return_value="dummy_tag")
        self.addCleanup(delattr, self, "fire_event_mock")
        opts = self.get_temp_config("master")

        patches = (
            ("zmq.Context", MagicMock()),
            ("salt.payload.Serial.dumps", MagicMock()),
            ("salt.master.tagify", MagicMock()),
            ("salt.utils.event.SaltEvent.fire_event", self.fire_event_mock),
            ("salt.auth.LoadAuth.time_auth", MagicMock(return_value=True)),
            ("salt.minion.MasterMinion", MagicMock()),
            ("salt.utils.verify.check_path_traversal", MagicMock()),
            ("salt.client.get_local_client", MagicMock()),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

        opts["publisher_acl"] = {}
        opts["publisher_acl_blacklist"] = {}
        opts["master_job_cache"] = ""
        opts["sign_pub_messages"] = False
        opts["con_cache"] = ""
        opts["external_auth"] = {}
        opts["external_auth"]["pam"] = {
            "test_user": [
                {"*": ["test.ping"]},
                {"minion_glob*": ["foo.bar"]},
                {"minion_func_test": ["func_test.*"]},
            ],
            "test_group%": [{"*": ["test.echo"]}],
            "test_user_mminion": [{"target_minion": ["test.ping"]}],
            "*": [{"my_minion": ["my_mod.my_func"]}],
            "test_user_func": [
                {
                    "*": [
                        {"test.echo": {"args": ["MSG:.*"]}},
                        {
                            "test.echo": {
                                "kwargs": {
                                    "text": "KWMSG:.*",
                                    "anything": ".*",
                                    "none": None,
                                }
                            }
                        },
                        {
                            "my_mod.*": {
                                "args": ["a.*", "b.*"],
                                "kwargs": {"kwa": "kwa.*", "kwb": "kwb"},
                            }
                        },
                    ]
                },
                {
                    "minion1": [
                        {"test.echo": {"args": ["TEST", None, "TEST.*"]}},
                        {"test.empty": {}},
                    ]
                },
            ],
        }
        self.clear = salt.master.ClearFuncs(opts, MagicMock())
        self.addCleanup(self.clear.destroy)
        self.addCleanup(delattr, self, "clear")

        # overwrite the _send_pub method so we don't have to serialize MagicMock
        self.clear._send_pub = lambda payload: True

        # make sure to return a JID, instead of a mock
        self.clear.mminion.returners = {".prep_jid": lambda x: 1}

        self.valid_clear_load = {
            "tgt_type": "glob",
            "jid": "",
            "cmd": "publish",
            "tgt": "test_minion",
            "kwargs": {
                "username": "test_user",
                "password": "test_password",
                "show_timeout": False,
                "eauth": "pam",
                "show_jid": False,
            },
            "ret": "",
            "user": "test_user",
            "key": "",
            "arg": "",
            "fun": "test.ping",
        }
        self.addCleanup(delattr, self, "valid_clear_load")

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_master_publish_name(self):
        """
        Test to ensure a simple name can auth against a given function.
        This tests to ensure test_user can access test.ping but *not* sys.doc
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            # Can we access test.ping?
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.ping")

            # Are we denied access to sys.doc?
            sys_doc_load = self.valid_clear_load
            sys_doc_load["fun"] = "sys.doc"
            self.clear.publish(sys_doc_load)
            self.assertIn(
                "error", self.fire_event_mock.call_args[0][0]
            )  # If sys.doc were to fire, this would match

    def test_master_publish_group(self):
        """
        Tests to ensure test_group can access test.echo but *not* sys.doc
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"]["user"] = "new_user"
            self.valid_clear_load["fun"] = "test.echo"
            self.valid_clear_load["arg"] = "hello"
            with patch(
                "salt.auth.LoadAuth.get_groups",
                return_value=["test_group", "second_test_group"],
            ):
                self.clear.publish(self.valid_clear_load)
            # Did we fire test.echo?
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.echo")

            # Request sys.doc
            self.valid_clear_load["fun"] = "sys.doc"
            # Did we fire it?
            self.assertNotEqual(self.fire_event_mock.call_args[0][0]["fun"], "sys.doc")

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_master_publish_some_minions(self):
        """
        Tests to ensure we can only target minions for which we
        have permission with publisher acl.

        Note that in order for these sorts of tests to run correctly that
        you should NOT patch check_minions!
        """
        self.valid_clear_load["kwargs"]["username"] = "test_user_mminion"
        self.valid_clear_load["user"] = "test_user_mminion"
        self.clear.publish(self.valid_clear_load)
        self.assertEqual(self.fire_event_mock.mock_calls, [])

    def test_master_not_user_glob_all(self):
        """
        Test to ensure that we DO NOT access to a given
        function to all users with publisher acl. ex:

        '*':
            my_minion:
                - my_func

        Yes, this seems like a bit of a no-op test but it's
        here to document that this functionality
        is NOT supported currently.

        WARNING: Do not patch this wit
        """
        self.valid_clear_load["kwargs"]["username"] = "NOT_A_VALID_USERNAME"
        self.valid_clear_load["user"] = "NOT_A_VALID_USERNAME"
        self.valid_clear_load["fun"] = "test.ping"
        self.clear.publish(self.valid_clear_load)
        self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
        self.assertEqual(self.fire_event_mock.mock_calls, [])

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_master_minion_glob(self):
        """
        Test to ensure we can allow access to a given
        function for a user to a subset of minions
        selected by a glob. ex:

        test_user:
            'minion_glob*':
              - glob_mod.glob_func

        This test is a bit tricky, because ultimately the real functionality
        lies in what's returned from check_minions, but this checks a limited
        amount of logic on the way there as well. Note the inline patch.
        """
        requested_function = "foo.bar"
        requested_tgt = "minion_glob1"
        self.valid_clear_load["tgt"] = requested_tgt
        self.valid_clear_load["fun"] = requested_function
        _check_minions_return = {"minions": ["minion_glob1"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):  # Assume that there is a listening minion match
            self.clear.publish(self.valid_clear_load)
        self.assertTrue(
            self.fire_event_mock.called,
            "Did not fire {} for minion tgt {}".format(
                requested_function, requested_tgt
            ),
        )
        self.assertEqual(
            self.fire_event_mock.call_args[0][0]["fun"],
            requested_function,
            f"Did not fire {requested_function} for minion glob",
        )

    def test_master_function_glob(self):
        """
        Test to ensure that we can allow access to a given
        set of functions in an execution module as selected
        by a glob. ex:

        my_user:
            my_minion:
                'test.*'
        """
        # Unimplemented

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_args_empty_spec(self):
        """
        Test simple arg restriction allowed.

        'test_user_func':
            minion1:
                - test.empty:
        """
        _check_minions_return = {"minions": ["minion1"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "minion1",
                    "fun": "test.empty",
                    "arg": ["TEST"],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.empty")

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_args_simple_match(self):
        """
        Test simple arg restriction allowed.

        'test_user_func':
            minion1:
                - test.echo:
                    args:
                        - 'TEST'
                        - 'TEST.*'
        """
        _check_minions_return = {"minions": ["minion1"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "minion1",
                    "fun": "test.echo",
                    "arg": ["TEST", "any", "TEST ABC"],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.echo")

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_args_more_args(self):
        """
        Test simple arg restriction allowed to pass unlisted args.

        'test_user_func':
            minion1:
                - test.echo:
                    args:
                        - 'TEST'
                        - 'TEST.*'
        """
        _check_minions_return = {"minions": ["minion1"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "minion1",
                    "fun": "test.echo",
                    "arg": [
                        "TEST",
                        "any",
                        "TEST ABC",
                        "arg 3",
                        {"kwarg1": "val1", "__kwarg__": True},
                    ],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.echo")

    def test_args_simple_forbidden(self):
        """
        Test simple arg restriction forbidden.

        'test_user_func':
            minion1:
                - test.echo:
                    args:
                        - 'TEST'
                        - 'TEST.*'
        """
        _check_minions_return = {"minions": ["minion1"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            # Wrong last arg
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "minion1",
                    "fun": "test.echo",
                    "arg": ["TEST", "any", "TESLA"],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Wrong first arg
            self.valid_clear_load["arg"] = ["TES", "any", "TEST1234"]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Missing the last arg
            self.valid_clear_load["arg"] = ["TEST", "any"]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # No args
            self.valid_clear_load["arg"] = []
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_args_kwargs_match(self):
        """
        Test simple kwargs restriction allowed.

        'test_user_func':
            '*':
                - test.echo:
                    kwargs:
                        text: 'KWMSG:.*'
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "*",
                    "fun": "test.echo",
                    "arg": [
                        {
                            "text": "KWMSG: a message",
                            "anything": "hello all",
                            "none": "hello none",
                            "__kwarg__": True,
                        }
                    ],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(self.fire_event_mock.call_args[0][0]["fun"], "test.echo")

    def test_args_kwargs_mismatch(self):
        """
        Test simple kwargs restriction allowed.

        'test_user_func':
            '*':
                - test.echo:
                    kwargs:
                        text: 'KWMSG:.*'
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {"user": "test_user_func", "tgt": "*", "fun": "test.echo"}
            )
            # Wrong kwarg value
            self.valid_clear_load["arg"] = [
                {
                    "text": "KWMSG a message",
                    "anything": "hello all",
                    "none": "hello none",
                    "__kwarg__": True,
                }
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Missing kwarg value
            self.valid_clear_load["arg"] = [
                {"anything": "hello all", "none": "hello none", "__kwarg__": True}
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            self.valid_clear_load["arg"] = [{"__kwarg__": True}]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            self.valid_clear_load["arg"] = [{}]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            self.valid_clear_load["arg"] = []
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Missing kwarg allowing any value
            self.valid_clear_load["arg"] = [
                {"text": "KWMSG: a message", "none": "hello none", "__kwarg__": True}
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            self.valid_clear_load["arg"] = [
                {"text": "KWMSG: a message", "anything": "hello all", "__kwarg__": True}
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_args_mixed_match(self):
        """
        Test mixed args and kwargs restriction allowed.

        'test_user_func':
            '*':
                - 'my_mod.*':
                    args:
                        - 'a.*'
                        - 'b.*'
                    kwargs:
                        'kwa': 'kwa.*'
                        'kwb': 'kwb'
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {
                    "user": "test_user_func",
                    "tgt": "*",
                    "fun": "my_mod.some_func",
                    "arg": [
                        "alpha",
                        "beta",
                        "gamma",
                        {
                            "kwa": "kwarg #1",
                            "kwb": "kwb",
                            "one_more": "just one more",
                            "__kwarg__": True,
                        },
                    ],
                }
            )
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(
                self.fire_event_mock.call_args[0][0]["fun"], "my_mod.some_func"
            )

    def test_args_mixed_mismatch(self):
        """
        Test mixed args and kwargs restriction forbidden.

        'test_user_func':
            '*':
                - 'my_mod.*':
                    args:
                        - 'a.*'
                        - 'b.*'
                    kwargs:
                        'kwa': 'kwa.*'
                        'kwb': 'kwb'
        """
        _check_minions_return = {"minions": ["some_minions"], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_check_minions_return),
        ):
            self.valid_clear_load["kwargs"].update({"username": "test_user_func"})
            self.valid_clear_load.update(
                {"user": "test_user_func", "tgt": "*", "fun": "my_mod.some_func"}
            )
            # Wrong arg value
            self.valid_clear_load["arg"] = [
                "alpha",
                "gamma",
                {
                    "kwa": "kwarg #1",
                    "kwb": "kwb",
                    "one_more": "just one more",
                    "__kwarg__": True,
                },
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Wrong kwarg value
            self.valid_clear_load["arg"] = [
                "alpha",
                "beta",
                "gamma",
                {
                    "kwa": "kkk",
                    "kwb": "kwb",
                    "one_more": "just one more",
                    "__kwarg__": True,
                },
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Missing arg
            self.valid_clear_load["arg"] = [
                "alpha",
                {
                    "kwa": "kwarg #1",
                    "kwb": "kwb",
                    "one_more": "just one more",
                    "__kwarg__": True,
                },
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])
            # Missing kwarg
            self.valid_clear_load["arg"] = [
                "alpha",
                "beta",
                "gamma",
                {"kwa": "kwarg #1", "one_more": "just one more", "__kwarg__": True},
            ]
            self.clear.publish(self.valid_clear_load)
            self.assertIn("error", self.fire_event_mock.mock_calls.pop(-1).args[0])
            self.assertEqual(self.fire_event_mock.mock_calls, [])


class AuthACLTestCase(ModuleCase):
    """
    A class to check various aspects of the publisher ACL system
    """

    def setUp(self):
        self.auth_check_mock = MagicMock(return_value=True)
        opts = self.get_temp_config("master")

        patches = (
            ("salt.minion.MasterMinion", MagicMock()),
            ("salt.utils.verify.check_path_traversal", MagicMock()),
            ("salt.utils.minions.CkMinions.auth_check", self.auth_check_mock),
            ("salt.auth.LoadAuth.time_auth", MagicMock(return_value=True)),
            ("salt.client.get_local_client", MagicMock()),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)
        self.addCleanup(delattr, self, "auth_check_mock")

        opts["publisher_acl"] = {}
        opts["publisher_acl_blacklist"] = {}
        opts["master_job_cache"] = ""
        opts["sign_pub_messages"] = False
        opts["con_cache"] = ""
        opts["external_auth"] = {}
        opts["external_auth"]["pam"] = {"test_user": [{"alpha_minion": ["test.ping"]}]}

        self.clear = salt.master.ClearFuncs(opts, MagicMock())
        self.addCleanup(self.clear.destroy)
        self.addCleanup(delattr, self, "clear")

        # overwrite the _send_pub method so we don't have to serialize MagicMock
        self.clear._send_pub = lambda payload: True

        # make sure to return a JID, instead of a mock
        self.clear.mminion.returners = {".prep_jid": lambda x: 1}

        self.valid_clear_load = {
            "tgt_type": "glob",
            "jid": "",
            "cmd": "publish",
            "tgt": "test_minion",
            "kwargs": {
                "username": "test_user",
                "password": "test_password",
                "show_timeout": False,
                "eauth": "pam",
                "show_jid": False,
            },
            "ret": "",
            "user": "test_user",
            "key": "",
            "arg": "",
            "fun": "test.ping",
        }
        self.addCleanup(delattr, self, "valid_clear_load")

    @pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
    def test_acl_simple_allow(self):
        self.clear.publish(self.valid_clear_load)
        self.assertEqual(
            self.auth_check_mock.call_args[0][0], [{"alpha_minion": ["test.ping"]}]
        )

    def test_acl_simple_deny(self):
        with patch(
            "salt.auth.LoadAuth.get_auth_list",
            MagicMock(return_value=[{"beta_minion": ["test.ping"]}]),
        ):
            self.clear.publish(self.valid_clear_load)
            self.assertEqual(
                self.auth_check_mock.call_args[0][0], [{"beta_minion": ["test.ping"]}]
            )



import salt.utils.pkg
from salt.utils.pkg import rpm
from tests.support.mock import ANY, MagicMock, patch
from tests.support.unit import TestCase


class PkgUtilsTestCase(TestCase):
    """
    TestCase for salt.utils.pkg module
    """

    test_parameters = [
        ("16.0.0.49153-0+f1", "", "16.0.0.49153-0+f1"),
        ("> 15.0.0", ">", "15.0.0"),
        ("< 15.0.0", "<", "15.0.0"),
        ("<< 15.0.0", "<<", "15.0.0"),
        (">> 15.0.0", ">>", "15.0.0"),
        (">= 15.0.0", ">=", "15.0.0"),
        ("<= 15.0.0", "<=", "15.0.0"),
        ("!= 15.0.0", "!=", "15.0.0"),
        ("<=> 15.0.0", "<=>", "15.0.0"),
        ("<> 15.0.0", "<>", "15.0.0"),
        ("= 15.0.0", "=", "15.0.0"),
        (">15.0.0", ">", "15.0.0"),
        ("<15.0.0", "<", "15.0.0"),
        ("<<15.0.0", "<<", "15.0.0"),
        (">>15.0.0", ">>", "15.0.0"),
        (">=15.0.0", ">=", "15.0.0"),
        ("<=15.0.0", "<=", "15.0.0"),
        ("!=15.0.0", "!=", "15.0.0"),
        ("<=>15.0.0", "<=>", "15.0.0"),
        ("<>15.0.0", "<>", "15.0.0"),
        ("=15.0.0", "=", "15.0.0"),
        ("", "", ""),
    ]

    def test_split_comparison(self):
        """
        Tests salt.utils.pkg.split_comparison
        """
        for test_parameter in self.test_parameters:
            oper, verstr = salt.utils.pkg.split_comparison(test_parameter[0])
            self.assertEqual(test_parameter[1], oper)
            self.assertEqual(test_parameter[2], verstr)


class PkgRPMTestCase(TestCase):
    """
    Test case for pkg.rpm utils
    """

    @patch("salt.utils.path.which", MagicMock(return_value=True))
    def test_get_osarch_by_rpm(self):
        """
        Get os_arch if RPM package is installed.
        :return:
        """
        subprocess_mock = MagicMock()
        subprocess_mock.Popen = MagicMock()
        subprocess_mock.Popen().communicate = MagicMock(return_value=["Z80"])
        with patch("salt.utils.pkg.rpm.subprocess", subprocess_mock):
            assert rpm.get_osarch() == "Z80"
        assert subprocess_mock.Popen.call_count == 2  # One within the mock
        subprocess_mock.Popen.assert_called_with(
            ["rpm", "--eval", "%{_host_cpu}"], close_fds=True, stderr=ANY, stdout=ANY
        )

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("salt.utils.pkg.rpm.subprocess", MagicMock(return_value=False))
    @patch(
        "salt.utils.pkg.rpm.platform.uname",
        MagicMock(
            return_value=(
                "Sinclair BASIC",
                "motophone",
                "1982 Sinclair Research Ltd",
                "1.0",
                "ZX81",
                "Z80",
            )
        ),
    )
    def test_get_osarch_by_platform(self):
        """
        Get os_arch if RPM package is not installed (inird image, for example).
        :return:
        """
        assert rpm.get_osarch() == "Z80"

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("salt.utils.pkg.rpm.subprocess", MagicMock(return_value=False))
    @patch(
        "salt.utils.pkg.rpm.platform.uname",
        MagicMock(
            return_value=(
                "Sinclair BASIC",
                "motophone",
                "1982 Sinclair Research Ltd",
                "1.0",
                "ZX81",
                "",
            )
        ),
    )
    def test_get_osarch_by_platform_no_cpu_arch(self):
        """
        Get os_arch if RPM package is not installed (inird image, for example) but cpu arch cannot be determined.
        :return:
        """
        assert rpm.get_osarch() == "ZX81"

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("salt.utils.pkg.rpm.subprocess", MagicMock(return_value=False))
    @patch(
        "salt.utils.pkg.rpm.platform.uname",
        MagicMock(
            return_value=(
                "Sinclair BASIC",
                "motophone",
                "1982 Sinclair Research Ltd",
                "1.0",
                "",
                "",
            )
        ),
    )
    def test_get_osarch_by_platform_no_cpu_arch_no_machine(self):
        """
        Get os_arch if RPM package is not installed (inird image, for example)
        where both cpu arch and machine cannot be determined.
        :return:
        """
        assert rpm.get_osarch() == "unknown"



"""
 Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values

  Originally posted at http://stackoverflow.com/questions/1165352/fast-comparison-between-two-python-dictionary/1165552#1165552
  Available at repository: https://github.com/hughdbrown/dictdiffer

  Added the ability to recursively compare dictionaries
"""

import copy
from collections.abc import Mapping


def diff(current_dict, past_dict):
    return DictDiffer(current_dict, past_dict)


class DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict, past_dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = set(list(current_dict)), set(list(past_dict))
        self.intersect = self.set_current.intersection(self.set_past)

    def added(self):
        return self.set_current - self.intersect

    def removed(self):
        return self.set_past - self.intersect

    def changed(self):
        return {o for o in self.intersect if self.past_dict[o] != self.current_dict[o]}

    def unchanged(self):
        return {o for o in self.intersect if self.past_dict[o] == self.current_dict[o]}


def deep_diff(old, new, ignore=None):
    ignore = ignore or []
    res = {}
    old = copy.deepcopy(old)
    new = copy.deepcopy(new)
    stack = [(old, new, False)]

    while stack:
        tmps = []
        tmp_old, tmp_new, reentrant = stack.pop()
        for key in set(list(tmp_old) + list(tmp_new)):
            if key in tmp_old and key in tmp_new and tmp_old[key] == tmp_new[key]:
                del tmp_old[key]
                del tmp_new[key]
                continue
            if not reentrant:
                if key in tmp_old and key in ignore:
                    del tmp_old[key]
                if key in tmp_new and key in ignore:
                    del tmp_new[key]
                if isinstance(tmp_old.get(key), Mapping) and isinstance(
                    tmp_new.get(key), Mapping
                ):
                    tmps.append((tmp_old[key], tmp_new[key], False))
        if tmps:
            stack.extend([(tmp_old, tmp_new, True)] + tmps)
    if old:
        res["old"] = old
    if new:
        res["new"] = new
    return res


def recursive_diff(past_dict, current_dict, ignore_missing_keys=True):
    """
    Returns a RecursiveDictDiffer object that computes the recursive diffs
    between two dictionaries

    past_dict
            Past dictionary

    current_dict
        Current dictionary

    ignore_missing_keys
        Flag specifying whether to ignore keys that no longer exist in the
        current_dict, but exist in the past_dict. If true, the diff will
        not contain the missing keys.
        Default is True.
    """
    return RecursiveDictDiffer(past_dict, current_dict, ignore_missing_keys)


class RecursiveDictDiffer(DictDiffer):
    """
    Calculates a recursive diff between the current_dict and the past_dict
    creating a diff in the format

    {'new': new_value, 'old': old_value}

    It recursively searches differences in common keys whose values are
    dictionaries creating a diff dict in the format

    {'common_key' : {'new': new_value, 'old': old_value}

    The class overrides all DictDiffer methods, returning lists of keys and
    subkeys using the . notation (i.e 'common_key1.common_key2.changed_key')

    The class provides access to:
        (1) the added, removed, changes keys and subkeys (using the . notation)
               ``added``, ``removed``, ``changed`` methods
        (2) the diffs in the format above (diff property)
                ``diffs`` property
        (3) a dict with the new changed values only (new_values property)
                ``new_values`` property
        (4) a dict with the old changed values only (old_values property)
                ``old_values`` property
        (5) a string representation of the changes in the format:
                ``changes_str`` property

    Note:
        The <_null_> value is a reserved value

    .. code-block:: text

        common_key1:
          common_key2:
            changed_key1 from '<old_str>' to '<new_str>'
            changed_key2 from '[<old_elem1>, ..]' to '[<new_elem1>, ..]'
        common_key3:
          changed_key3 from <old_int> to <new_int>

    """

    NONE_VALUE = "<_null_>"

    def __init__(self, past_dict, current_dict, ignore_missing_keys):
        """
        past_dict
            Past dictionary.

        current_dict
            Current dictionary.

        ignore_missing_keys
            Flag specifying whether to ignore keys that no longer exist in the
            current_dict, but exist in the past_dict. If true, the diff will
            not contain the missing keys.
        """
        super().__init__(current_dict, past_dict)
        self._diffs = self._get_diffs(
            self.current_dict, self.past_dict, ignore_missing_keys
        )
        # Ignores unet values when assessing the changes
        self.ignore_unset_values = True

    @classmethod
    def _get_diffs(cls, dict1, dict2, ignore_missing_keys):
        """
        Returns a dict with the differences between dict1 and dict2

        Notes:
            Keys that only exist in dict2 are not included in the diff if
            ignore_missing_keys is True, otherwise they are
            Simple compares are done on lists
        """
        ret_dict = {}
        for p in dict1:
            if p not in dict2:
                ret_dict.update({p: {"new": dict1[p], "old": cls.NONE_VALUE}})
            elif dict1[p] != dict2[p]:
                if isinstance(dict1[p], dict) and isinstance(dict2[p], dict):
                    sub_diff_dict = cls._get_diffs(
                        dict1[p], dict2[p], ignore_missing_keys
                    )
                    if sub_diff_dict:
                        ret_dict.update({p: sub_diff_dict})
                else:
                    ret_dict.update({p: {"new": dict1[p], "old": dict2[p]}})
        if not ignore_missing_keys:
            for p in dict2:
                if p not in dict1:
                    ret_dict.update({p: {"new": cls.NONE_VALUE, "old": dict2[p]}})
        return ret_dict

    @classmethod
    def _get_values(cls, diff_dict, type="new"):
        """
        Returns a dictionaries with the 'new' values in a diff dict.

        type
            Which values to return, 'new' or 'old'
        """
        ret_dict = {}
        for p in diff_dict:
            if type in diff_dict[p]:
                ret_dict.update({p: diff_dict[p][type]})
            else:
                ret_dict.update({p: cls._get_values(diff_dict[p], type=type)})
        return ret_dict

    @classmethod
    def _get_changes(cls, diff_dict):
        """
        Returns a list of string message with the differences in a diff dict.

        Each inner difference is tabulated two space deeper
        """
        changes_strings = []
        for p in sorted(diff_dict):
            if set(diff_dict[p]) == {"new", "old"}:
                changes = {
                    "old_value": diff_dict[p]["old"],
                    "new_value": diff_dict[p]["new"],
                }
                for ref in ("old_value", "new_value"):
                    val = changes[ref]
                    # Some string formatting
                    if val == cls.NONE_VALUE:
                        changes[ref] = "nothing"
                    elif isinstance(val, str):
                        changes[ref] = f"'{val}'"
                    elif isinstance(val, list):
                        changes[ref] = f"'{', '.join(val)}'"
                changes_strings.append(
                    f"{p} from {changes['old_value']} to {changes['new_value']}"
                )
            else:
                sub_changes = cls._get_changes(diff_dict[p])
                if sub_changes:
                    changes_strings.append(f"{p}:")
                    changes_strings.extend([f"  {c}" for c in sub_changes])
        return changes_strings

    def _it_addrm(
        self,
        key_a,
        key_b,
        include_nested=False,
        diffs=None,
        prefix="",
        is_nested=False,
        separator=".",
    ):
        keys = []
        if diffs is None:
            diffs = self.diffs

        for key in diffs:
            if is_nested:
                keys.append(f"{prefix}{key}")

            if not isinstance(diffs[key], dict):
                continue

            if is_nested and include_nested:
                keys.extend(
                    self._it_addrm(
                        key_a,
                        key_b,
                        diffs=diffs[key],
                        prefix=f"{prefix}{key}{separator}",
                        is_nested=is_nested,
                        include_nested=include_nested,
                        separator=separator,
                    )
                )
            elif "old" not in diffs[key]:
                keys.extend(
                    self._it_addrm(
                        key_a,
                        key_b,
                        diffs=diffs[key],
                        prefix=f"{prefix}{key}{separator}",
                        is_nested=is_nested,
                        include_nested=include_nested,
                        separator=separator,
                    )
                )
            elif diffs[key][key_a] == self.NONE_VALUE:
                keys.append(f"{prefix}{key}")

                if isinstance(diffs[key][key_b], dict) and include_nested:
                    keys.extend(
                        self._it_addrm(
                            key_a,
                            key_b,
                            diffs=diffs[key][key_b],
                            is_nested=True,
                            prefix=f"{prefix}{key}{separator}",
                            include_nested=include_nested,
                            separator=separator,
                        )
                    )
            # type change from/to dict
            elif (
                not is_nested
                and not isinstance(diffs[key][key_a], dict)
                and isinstance(diffs[key][key_b], dict)
            ):
                keys.extend(
                    self._it_addrm(
                        key_a,
                        key_b,
                        diffs=diffs[key][key_b],
                        is_nested=True,
                        prefix=f"{prefix}{key}{separator}",
                        include_nested=include_nested,
                        separator=separator,
                    )
                )

        return keys

    def added(
        self, include_nested=False, separator="."
    ):  # pylint: disable=arguments-differ
        """
        Returns all keys that have been added.

        include_nested
            If an added key contains a dictionary, include its
            keys in dot notation as well. Defaults to false.

            .. versionadded:: 3006.0

        separator
            Separator used to indicate nested keys. Defaults to ``.``.

            .. versionadded:: 3006.0
        """
        return sorted(self._it_addrm("old", "new", include_nested, separator=separator))

    def removed(
        self, include_nested=False, separator="."
    ):  # pylint: disable=arguments-differ
        """
        Returns all keys that have been removed.

        include_nested
            If an added key contains a dictionary, include its
            keys in dot notation as well. Defaults to false.

            .. versionadded:: 3006.0

        separator
            Separator used to indicate nested keys. Defaults to ``.``.

            .. versionadded:: 3006.0
        """
        return sorted(self._it_addrm("new", "old", include_nested, separator=separator))

    def changed(self, separator="."):  # pylint: disable=arguments-differ
        """
        Returns all keys that have been changed.

        separator
            Separator used to indicate nested keys. Defaults to ``.``.

            .. versionadded:: 3006.0
        """

        def _changed(diffs, prefix, separator):
            keys = []
            for key in diffs:
                if not isinstance(diffs[key], dict):
                    continue

                if isinstance(diffs[key], dict) and "old" not in diffs[key]:
                    keys.extend(
                        _changed(
                            diffs[key],
                            prefix=f"{prefix}{key}{separator}",
                            separator=separator,
                        )
                    )
                    continue
                if self.ignore_unset_values:
                    if (
                        "old" in diffs[key]
                        and "new" in diffs[key]
                        and diffs[key]["old"] != self.NONE_VALUE
                        and diffs[key]["new"] != self.NONE_VALUE
                    ):
                        if isinstance(diffs[key]["new"], dict):
                            keys.extend(
                                _changed(
                                    diffs[key]["new"],
                                    prefix=f"{prefix}{key}{separator}",
                                    separator=separator,
                                )
                            )
                        else:
                            keys.append(f"{prefix}{key}")
                    elif isinstance(diffs[key], dict):
                        keys.extend(
                            _changed(
                                diffs[key],
                                prefix=f"{prefix}{key}{separator}",
                                separator=separator,
                            )
                        )
                else:
                    if "old" in diffs[key] and "new" in diffs[key]:
                        if isinstance(diffs[key]["new"], dict):
                            keys.extend(
                                _changed(
                                    diffs[key]["new"],
                                    prefix=f"{prefix}{key}{separator}",
                                    separator=separator,
                                )
                            )
                        else:
                            keys.append(f"{prefix}{key}")
                    elif isinstance(diffs[key], dict):
                        keys.extend(
                            _changed(
                                diffs[key],
                                prefix=f"{prefix}{key}{separator}",
                                separator=separator,
                            )
                        )

            return keys

        return sorted(_changed(self._diffs, prefix="", separator=separator))

    def unchanged(self, separator="."):  # pylint: disable=arguments-differ
        """
        Returns all keys that have been unchanged.

        separator
            Separator used to indicate nested keys. Defaults to ``.``.

            .. versionadded:: 3006.0
        """

        def _unchanged(current_dict, diffs, prefix, separator):
            keys = []
            for key in current_dict:
                if key not in diffs:
                    keys.append(f"{prefix}{key}")
                elif isinstance(current_dict[key], dict):
                    if "new" in diffs[key]:
                        # There is a diff
                        continue
                    keys.extend(
                        _unchanged(
                            current_dict[key],
                            diffs[key],
                            prefix=f"{prefix}{key}{separator}",
                            separator=separator,
                        )
                    )

            return keys

        return sorted(
            _unchanged(self.current_dict, self._diffs, prefix="", separator=separator)
        )

    @property
    def diffs(self):
        """Returns a dict with the recursive diffs current_dict - past_dict"""
        return self._diffs

    @property
    def new_values(self):
        """Returns a dictionary with the new values"""
        return self._get_values(self._diffs, type="new")

    @property
    def old_values(self):
        """Returns a dictionary with the old values"""
        return self._get_values(self._diffs, type="old")

    @property
    def changes_str(self):
        """Returns a string describing the changes"""
        return "\n".join(self._get_changes(self._diffs))



import copy

import pytest

import salt.states.saltmod as saltmod
import salt.utils.state
from tests.support.mock import MagicMock, patch


class MockedEvent:
    """
    Mocked event class
    """

    def __init__(self, data):
        self.full = None
        self.flag = None
        self._data = data

    def get_event(self, full):
        """
        Mock get_event method
        """
        self.full = full
        if self.flag:
            return self._data
        return None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        saltmod: {
            "__opts__": minion_opts,
        },
    }


def test_test_mode():
    name = "presence"
    event_id = "lost"
    tgt = ["minion_1", "minion_2", "minion_3"]

    expected = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": f"Orchestration would wait for event '{name}'",
    }

    with patch.dict(saltmod.__opts__, {"test": True}):
        ret = saltmod.wait_for_event(name, tgt, event_id=event_id, timeout=-1.0)
        assert ret == expected


def test_wait_for_event():
    """
    Test to watch Salt's event bus and block until a condition is met
    """
    name = "state"
    tgt = "minion1"

    ret = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Timeout value reached.",
    }

    mocked_event = MockedEvent({"tag": name, "data": {}})
    with patch.object(
        salt.utils.event, "get_event", MagicMock(return_value=mocked_event)
    ):
        with patch.dict(saltmod.__opts__, {"sock_dir": True, "transport": True}):
            with patch("salt.states.saltmod.time.time", MagicMock(return_value=1.0)):
                assert saltmod.wait_for_event(name, "salt", timeout=-1.0) == ret

                mocked_event.flag = True
                ret.update(
                    {"comment": "All events seen in 0.0 seconds.", "result": True}
                )
                assert saltmod.wait_for_event(name, "") == ret

                ret.update({"comment": "Timeout value reached.", "result": False})
                assert saltmod.wait_for_event(name, tgt, timeout=-1.0) == ret


def test_list_single_event():
    """
    Test to watch Salt's event bus and block until a condition is met
    """
    name = "presence"
    event_id = "lost"
    tgt = ["minion_1", "minion_2", "minion_3"]

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Timeout value reached.",
    }

    mocked_event = MockedEvent({"tag": name, "data": {"lost": tgt}})
    with patch.object(
        salt.utils.event, "get_event", MagicMock(return_value=mocked_event)
    ):
        with patch.dict(saltmod.__opts__, {"sock_dir": True, "transport": True}):
            with patch("salt.states.saltmod.time.time", MagicMock(return_value=1.0)):
                expected.update({"comment": "Timeout value reached.", "result": False})
                ret = saltmod.wait_for_event(name, tgt, event_id=event_id, timeout=-1.0)
                assert ret == expected

                mocked_event.flag = True
                expected.update(
                    {
                        "name": name,
                        "changes": {"minions_seen": tgt},
                        "result": True,
                        "comment": "All events seen in 0.0 seconds.",
                    }
                )
                ret = saltmod.wait_for_event(
                    name, copy.deepcopy(tgt), event_id="lost", timeout=1.0
                )
                assert ret == expected



"""
Use Consul K/V as a Pillar source with values parsed as YAML

:depends:  - python-consul

In order to use an consul server, a profile must be created in the master
configuration file:

.. code-block:: yaml

    my_consul_config:
      consul.host: 127.0.0.1
      consul.port: 8500
      consul.token: b6376760-a8bb-edd5-fcda-33bc13bfc556
      consul.scheme: http
      consul.consistency: default
      consul.dc: dev
      consul.verify: True

All parameters are optional.

The ``consul.token`` requires python-consul >= 0.4.7.

If you have a multi-datacenter Consul cluster you can map your ``pillarenv``
entries to your data centers by providing a dictionary of mappings in
``consul.dc`` field:

.. code-block:: yaml

    my_consul_config:
      consul.dc:
        dev: us-east-1
        prod: us-west-1

In the example above we specifying static mapping between Pillar environments
and data centers: the data for ``dev`` and ``prod`` Pillar environments will
be fetched from ``us-east-1`` and ``us-west-1`` datacenter respectively.

In fact when ``consul.dc`` is set to dictionary keys are processed as regular
expressions (that can capture named parameters) and values are processed as
string templates as per PEP 3101.

.. code-block:: yaml

    my_consul_config:
      consul.dc:
        ^dev-.*$: dev-datacenter
        ^(?P<region>.*)-prod$: prod-datacenter-{region}

This example maps all Pillar environments starting with ``dev-`` to
``dev-datacenter`` whereas Pillar environment like ``eu-prod`` will be
mapped to ``prod-datacenter-eu``.

Before evaluation patterns are sorted by length in descending order.

If Pillar environment names correspond to data center names a single pattern
can be used:

.. code-block:: yaml

    my_consul_config:
      consul.dc:
        ^(?P<env>.*)$: '{env}'

After the profile is created, configure the external pillar system to use it.
Optionally, a root may be specified.

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config

    ext_pillar:
      - consul: my_consul_config root=salt

Using these configuration profiles, multiple consul sources may also be used:

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config
      - consul: my_other_consul_config

Either the ``minion_id``, or the ``role``, or the ``environment`` grain  may be used in the ``root``
path to expose minion-specific information stored in consul.

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config root=salt/%(minion_id)s
      - consul: my_consul_config root=salt/%(role)s
      - consul: my_consul_config root=salt/%(environment)s

Minion-specific values may override shared values when the minion-specific root
appears after the shared root:

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config root=salt-shared
      - consul: my_other_consul_config root=salt-private/%(minion_id)s

If using the ``role`` or ``environment`` grain in the consul key path, be sure to define it using
`/etc/salt/grains`, or similar:

.. code-block:: yaml

    role: my-minion-role
    environment: dev

It's possible to lock down where the pillar values are shared through minion
targeting. Note that double quotes ``"`` are required around the target value
and cannot be used inside the matching statement. See the section on Compound
Matchers for more examples.

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config root=salt target="L@salt.example.com and G@osarch:x86_64"

The data from Consul can be merged into a nested key in Pillar.

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config pillar_root=consul_data

By default, keys containing YAML data will be deserialized before being merged into Pillar.
This behavior can be disabled by setting ``expand_keys`` to ``false``.

.. code-block:: yaml

    ext_pillar:
      - consul: my_consul_config expand_keys=false

"""

import logging
import re

import salt.utils.minions
import salt.utils.yaml
from salt.exceptions import CommandExecutionError
from salt.utils.dictupdate import update as dict_merge

try:
    import consul

    if not hasattr(consul, "__version__"):
        consul.__version__ = (  # Some packages has no version, and so this pillar crashes on access to it.
            "0.1"
        )
except ImportError:
    consul = None

__virtualname__ = "consul"

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only return if python-consul is installed
    """
    return __virtualname__ if consul is not None else False


def ext_pillar(minion_id, pillar, conf):  # pylint: disable=W0613
    """
    Check consul for all data
    """
    opts = {}
    temp = conf
    target_re = re.compile('target="(.*?)"')
    match = target_re.search(temp)
    if match:
        opts["target"] = match.group(1)
        temp = temp.replace(match.group(0), "")
        checker = salt.utils.minions.CkMinions(__opts__)
        _res = checker.check_minions(opts["target"], "compound")
        minions = _res["minions"]
        log.debug("Targeted minions: %r", minions)
        if minion_id not in minions:
            return {}

    root_re = re.compile(r"(?<!_)root=(\S*)")  # pylint: disable=W1401
    match = root_re.search(temp)
    if match:
        opts["root"] = match.group(1).rstrip("/")
        temp = temp.replace(match.group(0), "")
    else:
        opts["root"] = ""

    pillar_root_re = re.compile(r"pillar_root=(\S*)")  # pylint: disable=W1401
    match = pillar_root_re.search(temp)
    if match:
        opts["pillar_root"] = match.group(1).rstrip("/")
        temp = temp.replace(match.group(0), "")
    else:
        opts["pillar_root"] = ""

    profile_re = re.compile(r"(?:profile=)?(\S+)")  # pylint: disable=W1401
    match = profile_re.search(temp)
    if match:
        opts["profile"] = match.group(1)
        temp = temp.replace(match.group(0), "")
    else:
        opts["profile"] = None

    expand_keys_re = re.compile(
        "expand_keys=False", re.IGNORECASE
    )  # pylint: disable=W1401
    match = expand_keys_re.search(temp)
    if match:
        opts["expand_keys"] = False
        temp = temp.replace(match.group(0), "")
    else:
        opts["expand_keys"] = True

    client = get_conn(__opts__, opts["profile"])

    role = __salt__["grains.get"]("role", None)
    environment = __salt__["grains.get"]("environment", None)
    # put the minion's ID in the path if necessary
    opts["root"] %= {"minion_id": minion_id, "role": role, "environment": environment}

    try:
        pillar_tree = fetch_tree(client, opts["root"], opts["expand_keys"])
        if opts["pillar_root"]:
            log.debug(
                "Merging consul path %s/ into pillar at %s/",
                opts["root"],
                opts["pillar_root"],
            )

            pillar = {}
            branch = pillar
            keys = opts["pillar_root"].split("/")

            for i, k in enumerate(keys):
                if i == len(keys) - 1:
                    branch[k] = pillar_tree
                else:
                    branch[k] = {}
                    branch = branch[k]
        else:
            pillar = pillar_tree
    except KeyError:
        log.error("No such key in consul profile %s: %s", opts["profile"], opts["root"])
        pillar = {}

    return pillar


def consul_fetch(client, path):
    """
    Query consul for all keys/values within base path
    """
    # Unless the root path is blank, it needs a trailing slash for
    # the kv get from Consul to work as expected
    return client.kv.get("" if not path else path.rstrip("/") + "/", recurse=True)


def fetch_tree(client, path, expand_keys):
    """
    Grab data from consul, trim base path and remove any keys which
    are folders. Take the remaining data and send it to be formatted
    in such a way as to be used as pillar data.
    """
    _, items = consul_fetch(client, path)
    ret = {}
    has_children = re.compile(r"/$")

    log.debug("Fetched items: %r", items)

    if items is None:
        return ret
    for item in reversed(items):
        key = re.sub(r"^" + re.escape(path) + "/?", "", item["Key"])
        if key != "":
            log.debug("path/key - %s: %s", path, key)
            log.debug("has_children? %r", has_children.search(key))
        if has_children.search(key) is None:
            ret = pillar_format(ret, key.split("/"), item["Value"], expand_keys)
            log.debug("Fetching subkeys for key: %r", item)

    return ret


def pillar_format(ret, keys, value, expand_keys):
    """
    Perform data formatting to be used as pillar data and
    merge it with the current pillar data
    """
    # if value is empty in Consul then it's None here - skip it
    if value is None:
        return ret

    # If value is not None then it's a string
    # YAML strips whitespaces unless they're surrounded by quotes
    # If expand_keys is true, deserialize the YAML data
    if expand_keys:
        pillar_value = salt.utils.yaml.safe_load(value)
    else:
        pillar_value = value

    keyvalue = keys.pop()
    pil = {keyvalue: pillar_value}
    keys.reverse()
    for k in keys:
        pil = {k: pil}

    return dict_merge(ret, pil)


def get_conn(opts, profile):
    """
    Return a client object for accessing consul
    """
    opts_pillar = opts.get("pillar", {})
    opts_master = opts_pillar.get("master", {})

    opts_merged = {}
    opts_merged.update(opts_master)
    opts_merged.update(opts_pillar)
    opts_merged.update(opts)

    if profile:
        conf = opts_merged.get(profile, {})
    else:
        conf = opts_merged

    params = {}
    for key in conf:
        if key.startswith("consul."):
            params[key.split(".")[1]] = conf[key]

    if "dc" in params:
        pillarenv = opts_merged.get("pillarenv") or "base"
        params["dc"] = _resolve_datacenter(params["dc"], pillarenv)

    if consul:
        # Sanity check. ACL Tokens are supported on python-consul 0.4.7 onwards only.
        if consul.__version__ < "0.4.7" and params.get("target"):
            params.pop("target")
        return consul.Consul(**params)
    else:
        raise CommandExecutionError(
            "(unable to import consul, "
            "module most likely not installed. Download python-consul "
            "module and be sure to import consul)"
        )


def _resolve_datacenter(dc, pillarenv):
    """
    If ``dc`` is a string - return it as is.

    If it's a dict then sort it in descending order by key length and try
    to use keys as RegEx patterns to match against ``pillarenv``.
    The value for matched pattern should be a string (that can use
    ``str.format`` syntax togetehr with captured variables from pattern)
    pointing to targe data center to use.

    If none patterns matched return ``None`` which meanse us datacenter of
    conencted Consul agent.
    """
    log.debug("Resolving Consul datacenter based on: %s", dc)

    try:
        mappings = dc.items()  # is it a dict?
    except AttributeError:
        log.debug("Using pre-defined DC: '%s'", dc)
        return dc

    log.debug("Selecting DC based on pillarenv using %d pattern(s)", len(mappings))
    log.debug("Pillarenv set to '%s'", pillarenv)

    # sort in reverse based on pattern length
    # but use alphabetic order within groups of patterns of same length
    sorted_mappings = sorted(mappings, key=lambda m: (-len(m[0]), m[0]))

    for pattern, target in sorted_mappings:
        match = re.match(pattern, pillarenv)
        if match:
            log.debug("Matched pattern: '%s'", pattern)
            result = target.format(**match.groupdict())
            log.debug("Resolved datacenter: '%s'", result)
            return result

    log.debug(
        "None of following patterns matched pillarenv=%s: %s",
        pillarenv,
        ", ".join(repr(x) for x in mappings),
    )
